from typing import Callable, Dict, List, Optional

import metatensor.torch
import numpy as np
import torch
from metatensor.torch import Labels, TensorBlock, TensorMap
from metatensor.torch.atomistic import (
    ModelCapabilities,
    ModelEvaluationOptions,
    ModelOutput,
    System,
)
from torch.utils.data import DataLoader

from metatrain.utils.data.target_info import is_auxiliary_output

from .data import DatasetInfo, TargetInfo, get_atomic_types
from .evaluate_model import evaluate_model
from .per_atom import average_by_num_atoms


class LLPRUncertaintyModel(torch.nn.Module):
    """A wrapper that adds LLPR uncertainties to a model.

    In order to be compatible with this class, a model needs to have the last-layer
    feature size available as an attribute (with the ``last_layer_feature_size`` name)
    and be capable of returning last-layer features (see auxiliary outputs in
    metatrain), optionally per atom to calculate LPRs with the LLPR method.

    :param model: The model to wrap.
    """

    def __init__(
        self,
        model: torch.jit._script.RecursiveScriptModule,
    ) -> None:
        super().__init__()

        self.model = model
        self.ll_feat_size = self.model.module.last_layer_feature_size

        # update capabilities: now we have additional outputs for the uncertainty
        old_capabilities = self.model.capabilities()
        additional_capabilities = {}
        self.uncertainty_multipliers = {}
        for name, output in old_capabilities.outputs.items():
            if is_auxiliary_output(name):
                continue  # auxiliary output
            uncertainty_name = f"mtt::aux::{name.replace('mtt::', '')}_uncertainty"
            additional_capabilities[uncertainty_name] = ModelOutput(
                quantity="",
                unit=f"({output.unit})^2",
                per_atom=True,
            )
            self.uncertainty_multipliers[uncertainty_name] = 1.0
        self.capabilities = ModelCapabilities(
            outputs={**old_capabilities.outputs, **additional_capabilities},
            atomic_types=old_capabilities.atomic_types,
            interaction_range=old_capabilities.interaction_range,
            length_unit=old_capabilities.length_unit,
            supported_devices=old_capabilities.supported_devices,
            dtype=old_capabilities.dtype,
        )

        # register covariance and inverse covariance buffers
        device = next(self.model.parameters()).device
        dtype = getattr(torch, old_capabilities.dtype)
        self.covariances = {
            uncertainty_name: torch.zeros(
                (self.ll_feat_size, self.ll_feat_size),
                device=device,
                dtype=dtype,
            )
            for uncertainty_name in self.uncertainty_multipliers.keys()
        }
        self.inv_covariances = {
            uncertainty_name: torch.zeros(
                (self.ll_feat_size, self.ll_feat_size),
                device=device,
                dtype=dtype,
            )
            for uncertainty_name in self.uncertainty_multipliers.keys()
        }

        # flags
        self.covariance_computed = False
        self.inv_covariance_computed = False
        self.is_calibrated = False

    def forward(
        self,
        systems: List[System],
        outputs: Dict[str, ModelOutput],
        selected_atoms: Optional[Labels] = None,
    ) -> Dict[str, TensorMap]:
        device = systems[0].positions.device
        if list(self.covariances.values())[0].device != device:
            for name in self.covariances.keys():
                self.covariances[name] = self.covariances[name].to(device=device)
                self.inv_covariances[name] = self.inv_covariances[name].to(
                    device=device
                )

        if not self.inv_covariance_computed:
            raise ValueError(
                "Trying to predict with LLPR, but inverse covariance has not "
                "been computed yet."
            )

        if all("_uncertainty" not in output for output in outputs):
            # no uncertainties requested
            options = ModelEvaluationOptions(
                length_unit="",
                outputs=outputs,
                selected_atoms=selected_atoms,
            )
            return self.model(systems, options, check_consistency=False)

        per_atom_all_targets = [output.per_atom for output in outputs.values()]
        # impose either all per atom or all not per atom
        if not all(per_atom_all_targets) and any(per_atom_all_targets):
            raise ValueError(
                "All output uncertainties must be either be requested per "
                "atom or not per atom with LLPR."
            )
        per_atom = per_atom_all_targets[0]
        outputs_for_model: Dict[str, ModelOutput] = {}
        for name in outputs.keys():
            if name.endswith("_uncertainty"):
                base_name = name.replace("_uncertainty", "").replace("mtt::aux::", "")
                if base_name not in outputs and f"mtt::{base_name}" not in outputs:
                    raise ValueError(
                        f"Requested uncertainty '{name}' without corresponding "
                        f"output `{base_name}` (or `mtt::{base_name}`)."
                    )
                # request corresponding features
                target_name = name.replace("mtt::aux::", "").replace("_uncertainty", "")
                outputs_for_model[f"mtt::aux::{target_name}_last_layer_features"] = (
                    ModelOutput(
                        quantity="",
                        unit="",
                        per_atom=per_atom,
                    )
                )
        for name, output in outputs.items():
            # remove uncertainties from the requested outputs for the
            # wrapped model
            if name.startswith("mtt::aux") and name.endswith("_uncertainty"):
                continue
            if name.endswith("_ensemble"):
                continue
            outputs_for_model[name] = output

        options = ModelEvaluationOptions(
            length_unit="",
            outputs=outputs_for_model,
            selected_atoms=selected_atoms,
        )
        return_dict = self.model(systems, options, check_consistency=False)

        requested_uncertainties: List[str] = []
        for name in outputs.keys():
            if name.startswith("mtt::aux::") and name.endswith("_uncertainty"):
                requested_uncertainties.append(name)

        for name in requested_uncertainties:
            ll_features = return_dict[
                name.replace("_uncertainty", "_last_layer_features")
            ]

            # compute PRs
            # the code is the same for PR and LPR
            one_over_pr_values = torch.einsum(
                "ij, jk, ik -> i",
                ll_features.block().values,
                self.inv_covariances[name],
                ll_features.block().values,
            ).unsqueeze(1)
            one_over_pr = TensorMap(
                keys=Labels(
                    names=["_"],
                    values=torch.tensor(
                        [[0]], device=ll_features.block().values.device
                    ),
                ),
                blocks=[
                    TensorBlock(
                        values=one_over_pr_values,
                        samples=ll_features.block().samples,
                        components=ll_features.block().components,
                        properties=Labels(
                            names=["_"],
                            values=torch.tensor(
                                [[0]], device=ll_features.block().values.device
                            ),
                        ),
                    )
                ],
            )

            return_dict[name] = metatensor.torch.multiply(
                one_over_pr, self.uncertainty_multipliers[name]
            )

        # now deal with potential ensembles (see generate_ensemble method)
        requested_ensembles: List[str] = []
        for name in outputs.keys():
            if name.endswith("_ensemble"):
                requested_ensembles.append(name)

        for name in requested_ensembles:
            ll_features_name = name.replace("_ensemble", "_last_layer_features")
            if ll_features_name == "energy_last_layer_features":
                # special case for energy_ensemble
                ll_features_name = "mtt::aux::energy_last_layer_features"
            ll_features = return_dict[ll_features_name]
            # get the ensemble weights (getattr not supported by torchscript)
            ensemble_weights = torch.tensor(0.0)
            for buffer_name, buffer in self.named_buffers():
                if buffer_name == name + "_weights":
                    ensemble_weights = buffer
            # the ensemble weights should always be found (checks are performed
            # in the generate_ensemble method and in the metatensor wrapper)
            ensemble_values = torch.einsum(
                "ij, jk -> ik",
                ll_features.block().values,
                ensemble_weights,
            )

            # since we know the exact mean of the ensemble from the model's prediction,
            # it should be mathematically correct to use it to re-center the ensemble.
            # Besides making sure that the average is always correct (so that results
            # will always be consistent between LLPR ensembles and the original model),
            # this also takes care of additive contributions that are not present in the
            # last layer, which can be composition, short-range models, a bias in the
            # last layer, etc.
            original_name = (
                name.replace("_ensemble", "").replace("aux::", "")
                if name.replace("_ensemble", "").replace("aux::", "") in outputs
                else name.replace("_ensemble", "").replace("mtt::aux::", "")
            )
            ensemble_values = (
                ensemble_values
                - ensemble_values.mean(dim=1, keepdim=True)
                + return_dict[original_name].block().values
            )

            property_name = "energy" if name == "energy_ensemble" else "ensemble_member"
            ensemble = TensorMap(
                keys=Labels(
                    names=["_"],
                    values=torch.tensor(
                        [[0]], device=ll_features.block().values.device
                    ),
                ),
                blocks=[
                    TensorBlock(
                        values=ensemble_values,
                        samples=ll_features.block().samples,
                        components=ll_features.block().components,
                        properties=Labels(
                            names=[property_name],
                            values=torch.arange(
                                ensemble_values.shape[1], device=ensemble_values.device
                            ).unsqueeze(1),
                        ),
                    )
                ],
            )
            return_dict[name] = ensemble

        # remove the last-layer features from return_dict if they were not requested
        for key in list(return_dict.keys()):
            if key.endswith("_last_layer_features"):
                if key not in outputs:
                    return_dict.pop(key)

        return return_dict

    def compute_covariance(self, train_loader: DataLoader) -> None:
        """A function to compute the covariance matrix for a training set.

        The covariance is stored as a buffer in the model.

        :param train_loader: A PyTorch DataLoader with the training data.
            The individual samples need to be compatible with the ``Dataset``
            class in ``metatrain``.
        """
        device = next(iter(self.covariances.values())).device
        dtype = next(iter(self.covariances.values())).dtype
        for batch in train_loader:
            systems, targets = batch
            n_atoms = torch.tensor(
                [len(system.positions) for system in systems], device=device
            )
            systems = [system.to(device=device, dtype=dtype) for system in systems]
            options_for_targets = {
                name: ModelOutput(
                    quantity="",
                    unit="",
                    per_atom=False,
                )
                for name in targets.keys()
            }
            outputs_for_features = {
                f"mtt::aux::{name.replace('mtt::', '')}_last"
                "_layer_features": ModelOutput(
                    quantity="",
                    unit="",
                    per_atom=False,
                )
                for name in targets.keys()
            }
            options = ModelEvaluationOptions(
                length_unit="",
                outputs={**options_for_targets, **outputs_for_features},
            )
            output = self.model(systems, options, check_consistency=False)
            for name in targets.keys():
                ll_feat_tmap = output[
                    f"mtt::aux::{name.replace('mtt::', '')}_last_layer_features"
                ]
                ll_feats = ll_feat_tmap.block().values.detach() / n_atoms.unsqueeze(1)
                self.covariances[
                    f"mtt::aux::{name.replace('mtt::', '')}_uncertainty"
                ] += ll_feats.T @ ll_feats

        self.covariance_computed = True

    def compute_covariance_as_pseudo_hessian(
        self,
        train_loader: DataLoader,
        target_infos: Dict[str, TargetInfo],
        loss_fn: Callable,
        parameters: Dict[str, List[torch.nn.Parameter]],
    ) -> None:
        """A function to compute the covariance matrix for a training set
        as the pseudo-Hessian of the loss function.

        The covariance/pseudo-Hessian is stored as a buffer in the model. The
        loss function must be compatible with the Dataloader (i.e., it should
        have the same structure as the outputs of the model and as the targets
        in the dataset). All contributions to the loss functions are assumed to
        be per-atom, except for quantities that are already per-atom (e.g.,
        forces).

        :param train_loader: A PyTorch DataLoader with the training data.
            The individual samples need to be compatible with the ``Dataset``
            class in ``metatrain``.
        :param loss_fn: A loss function that takes the model outputs and the
            targets and returns a scalar loss.
        :param parameters: A list of model parameters for which the pseudo-Hessian
            should be computed. This is often necessary as models can have very
            large numbers of parameters, and the pseudo-Hessian's number of
            elements grows quadratically with the number of parameters. For this
            reason, only a subset of the parameters of the model is usually used
            in the calculation. This list allows the user to feed the parameters
            of interest directly to the function. In order to function correctly,
            the model's parameters should be those corresponding to the last
            layer(s) of the model, such that their concatenation corresponds to the
            last-layer features, in the same order as those are returned by the
            base model.
        """
        self.model = self.model.train()  # we need gradients w.r.t. parameters
        # disable gradients for all parameters that are not in the list
        for parameter in self.model.parameters():
            parameter.requires_grad = False
        all_parameters_that_require_grad = []
        all_output_names = []
        for output_name, output_parameters in parameters.items():
            for parameter in output_parameters:
                all_parameters_that_require_grad.append(parameter)
                all_output_names.append(output_name)
                parameter.requires_grad = True

        dataset = train_loader.dataset
        dataset_info = DatasetInfo(
            length_unit=self.capabilities.length_unit,  # TODO: check
            atomic_types=get_atomic_types(dataset),
            targets=target_infos,
        )
        train_targets = dataset_info.targets
        device = next(iter(self.covariances.values())).device
        dtype = next(iter(self.covariances.values())).dtype
        for batch in train_loader:
            systems, targets = batch
            systems = [system.to(device=device, dtype=dtype) for system in systems]
            targets = {
                name: tmap.to(device=device, dtype=dtype)
                for name, tmap in targets.items()
            }
            predictions = evaluate_model(
                self.model,
                systems,
                {key: train_targets[key] for key in targets.keys()},
                is_training=True,  # keep the computational graph
            )

            # average by the number of atoms
            predictions = average_by_num_atoms(predictions, systems, [])
            targets = average_by_num_atoms(targets, systems, [])

            loss = loss_fn(predictions, targets)

            grads = torch.autograd.grad(
                loss,
                all_parameters_that_require_grad,
                create_graph=False,
                retain_graph=False,
                allow_unused=True,  # if there are multiple last-layers
                materialize_grads=True,  # avoid Nones
            )

            for output_name in np.unique(all_output_names):
                grads = [
                    grad
                    for grad, name in zip(grads, all_output_names)
                    if name == output_name
                ]
                grads = torch.cat(grads, dim=1)
                self.covariances[
                    "mtt::aux::" + output_name.replace("mtt::", "") + "_uncertainty"
                ] += grads.T @ grads

            for parameter in all_parameters_that_require_grad:
                parameter.grad = None  # reset the gradients

        self.covariance_computed = True

        for parameter in self.model.parameters():
            parameter.requires_grad = True

        self.model = self.model.eval()  # restore the model to evaluation mode

    def compute_inverse_covariance(self, regularizer: Optional[float] = None):
        """A function to compute the inverse covariance matrix.

        The inverse covariance is stored as a buffer in the model.

        :param regularizer: A regularization parameter to ensure the matrix is
            invertible. If not provided, the function will try to compute the
            inverse without regularization and increase the regularization
            parameter until the matrix is invertible.
        """
        for name in self.covariances.keys():
            if regularizer is not None:
                self.inv_covariances[name] = torch.inverse(
                    self.covariances[name]
                    + regularizer
                    * torch.eye(self.ll_feat_size, device=self.covariances[name].device)
                )
            else:
                # Try with an increasingly high regularization parameter until
                # the matrix is invertible
                def is_psd(x):
                    return torch.all(torch.linalg.eigvalsh(x) >= 0.0)

                for log10_sigma_squared in torch.linspace(-20.0, 16.0, 33):
                    if not is_psd(
                        self.covariances[name]
                        + 10**log10_sigma_squared
                        * torch.eye(
                            self.ll_feat_size, device=self.covariances[name].device
                        )
                    ):
                        continue
                    else:
                        inverse = torch.inverse(
                            self.covariances[name]
                            + 10 ** (log10_sigma_squared + 0.0)
                            * torch.eye(
                                self.ll_feat_size, device=self.covariances[name].device
                            )
                        )
                        self.inv_covariances[name] = (inverse + inverse.T) / 2.0
                        break

        self.inv_covariance_computed = True

    def calibrate(self, valid_loader: DataLoader):
        """
        Calibrate the LLPR model.

        This function computes the calibration constants (one for each output)
        that are used to scale the uncertainties in the LLPR model. The
        calibration is performed in a simple way by computing the calibration
        constant as the mean of the squared residuals divided by the mean of
        the non-calibrated uncertainties.

        :param valid_loader: A data loader with the validation data.
            This data loader should be generated from a dataset from the
            ``Dataset`` class in ``metatrain.utils.data``.
        """
        # calibrate the LLPR
        # TODO: in the future, we might want to have one calibration factor per
        # property for outputs with multiple properties
        device = next(iter(self.covariances.values())).device
        dtype = next(iter(self.covariances.values())).dtype
        all_predictions = {}  # type: ignore
        all_targets = {}  # type: ignore
        all_uncertainties = {}  # type: ignore
        for batch in valid_loader:
            systems, targets = batch
            systems = [system.to(device=device, dtype=dtype) for system in systems]
            targets = {
                name: target.to(device=device, dtype=dtype)
                for name, target in targets.items()
            }
            # evaluate the targets and their uncertainties, not per atom
            requested_outputs = {}
            for name in targets:
                requested_outputs[name] = ModelOutput(
                    quantity="",
                    unit="",
                    per_atom=False,
                )
                uncertainty_name = f"mtt::aux::{name.replace('mtt::', '')}_uncertainty"
                requested_outputs[uncertainty_name] = ModelOutput(
                    quantity="",
                    unit="",
                    per_atom=False,
                )
            outputs = self.forward(systems, requested_outputs)
            for name, target in targets.items():
                uncertainty_name = f"mtt::aux::{name.replace('mtt::', '')}_uncertainty"
                if name not in all_predictions:
                    all_predictions[name] = []
                    all_targets[name] = []
                    all_uncertainties[uncertainty_name] = []
                all_predictions[name].append(outputs[name].block().values.detach())
                all_targets[name].append(target.block().values)
                all_uncertainties[uncertainty_name].append(
                    outputs[uncertainty_name].block().values.detach()
                )

        for name in all_predictions:
            all_predictions[name] = torch.cat(all_predictions[name], dim=0)
            all_targets[name] = torch.cat(all_targets[name], dim=0)
            uncertainty_name = f"mtt::aux::{name.replace('mtt::', '')}_uncertainty"
            all_uncertainties[uncertainty_name] = torch.cat(
                all_uncertainties[uncertainty_name], dim=0
            )

        for name in all_predictions:
            # compute the uncertainty multiplier
            residuals = all_predictions[name] - all_targets[name]
            uncertainty_name = f"mtt::aux::{name.replace('mtt::', '')}_uncertainty"
            uncertainties = all_uncertainties[uncertainty_name]
            self.uncertainty_multipliers[uncertainty_name] = torch.mean(
                residuals**2 / uncertainties
            ).item()

        self.is_calibrated = True

    def generate_ensemble(
        self, weight_tensors: Dict[str, torch.Tensor], n_members: int
    ) -> None:
        """Generate an ensemble of weights for the model.

        The ensemble is generated by sampling from a multivariate normal
        distribution with mean given by the input weights and covariance given
        by the inverse covariance matrix.

        :param weight_tensors: A dictionary with the weights for the ensemble.
            The keys should be the names of the weights in the model and the
            values should be 1D PyTorch tensors.
        :param n_members: The number of members in the ensemble.
        """
        # note: we could also allow n_members to be different for each output

        # basic checks
        if not self.is_calibrated:
            raise ValueError(
                "LLPR model needs to be calibrated before generating ensembles"
            )
        for key in weight_tensors:
            if key not in self.capabilities.outputs.keys():
                raise ValueError(f"Output '{key}' not supported by model")
            if len(weight_tensors[key].shape) != 1:
                raise ValueError("All weights must be 1D tensors")

        # sampling; each member is sampled from a multivariate normal distribution
        # with mean given by the input weights and covariance given by the inverse
        # covariance matrix
        for name, weights in weight_tensors.items():
            uncertainty_name = "mtt::aux::" + name.replace("mtt::", "") + "_uncertainty"
            device = self.inv_covariances[uncertainty_name].device
            dtype = self.inv_covariances[uncertainty_name].dtype
            rng = np.random.default_rng()
            ensemble_weights = rng.multivariate_normal(
                weights.clone().detach().cpu().numpy(),
                self.inv_covariances[uncertainty_name].clone().detach().cpu().numpy()
                * self.uncertainty_multipliers[uncertainty_name],
                size=n_members,
                method="svd",
            ).T
            ensemble_weights = torch.tensor(
                ensemble_weights, device=device, dtype=dtype
            )
            ensemble_weights_name = (
                "mtt::aux::" + name.replace("mtt::", "") + "_ensemble_weights"
            )
            if ensemble_weights_name == "mtt::aux::energy_ensemble_weights":
                ensemble_weights_name = "energy_ensemble_weights"
            self.register_buffer(
                ensemble_weights_name,
                ensemble_weights,
            )

        # add the ensembles to the capabilities
        old_outputs = self.capabilities.outputs
        new_outputs = {}
        for name in weight_tensors.keys():
            ensemble_name = "mtt::aux::" + name.replace("mtt::", "") + "_ensemble"
            if ensemble_name == "mtt::aux::energy_ensemble":
                ensemble_name = "energy_ensemble"
            new_outputs[ensemble_name] = ModelOutput(
                quantity=old_outputs[name].quantity,
                unit=old_outputs[name].unit,
                per_atom=old_outputs[name].per_atom,
            )
        self.capabilities = ModelCapabilities(
            outputs={**old_outputs, **new_outputs},
            atomic_types=self.capabilities.atomic_types,
            interaction_range=self.capabilities.interaction_range,
            length_unit=self.capabilities.length_unit,
            supported_devices=self.capabilities.supported_devices,
            dtype=self.capabilities.dtype,
        )



import pytest
import torch
from metatensor.torch.atomistic import (
    ModelCapabilities,
    ModelEvaluationOptions,
    ModelMetadata,
    ModelOutput,
    System,
)

from metatrain.pet import PET as WrappedPET
from metatrain.pet.modules.hypers import Hypers
from metatrain.pet.modules.pet import PET
from metatrain.utils.architectures import get_default_hypers
from metatrain.utils.data import DatasetInfo
from metatrain.utils.data.target_info import get_energy_target_info
from metatrain.utils.neighbor_lists import (
    get_requested_neighbor_lists,
    get_system_with_neighbor_lists,
)


DEFAULT_HYPERS = get_default_hypers("pet")


@pytest.mark.parametrize("device", ["cpu", "cuda"])
def test_to(device):
    """Tests that the `.to()` method of the exported model works."""
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA is not available")

    dtype = torch.float32  # for now
    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )
    model = WrappedPET(DEFAULT_HYPERS["model"], dataset_info)
    ARCHITECTURAL_HYPERS = Hypers(model.hypers)
    raw_pet = PET(ARCHITECTURAL_HYPERS, 0.0, len(model.atomic_types))
    model.set_trained_model(raw_pet)

    capabilities = ModelCapabilities(
        length_unit="Angstrom",
        atomic_types=model.atomic_types,
        outputs={
            "energy": ModelOutput(
                quantity="energy",
                unit="eV",
            )
        },
        interaction_range=DEFAULT_HYPERS["model"]["N_GNN_LAYERS"]
        * DEFAULT_HYPERS["model"]["R_CUT"],
        dtype="float32",
        supported_devices=["cpu", "cuda"],
    )

    exported = model.export(metadata=ModelMetadata(name="test"))

    # test correct metadata
    assert "This is the test model" in str(exported.metadata())

    exported.to(device=device, dtype=dtype)

    system = System(
        types=torch.tensor([6, 6]),
        positions=torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]]),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    requested_neighbor_lists = get_requested_neighbor_lists(exported)
    system = get_system_with_neighbor_lists(system, requested_neighbor_lists)
    system = system.to(device=device, dtype=dtype)

    evaluation_options = ModelEvaluationOptions(
        length_unit=dataset_info.length_unit,
        outputs=capabilities.outputs,
    )

    exported([system], evaluation_options, check_consistency=True)



import json

from metatensor.torch.atomistic import ModelMetadata


def append_metadata_references(self: ModelMetadata, other: ModelMetadata) -> None:
    """Append ``references`` to an existing ModelMetadata object.

    :param self: The metadata object to be appeneded.
    :param other: The metadata object to update with.
    """

    self_dict = json.loads(self._get_method("__getstate__")())
    other_dict = json.loads(other._get_method("__getstate__")())

    for key, values in other_dict["references"].items():
        if key not in self_dict["references"]:
            self_dict["references"][key] = values
        else:
            self_dict["references"][key] += values

    self._get_method("__setstate__")(json.dumps(self_dict))



from typing import Optional

import torch


# NEF stands for Node Edge Feature, representing the three dimensions of
# the internal representations of a PET model. This module contains
# functions to manipulate these representations.


def get_nef_indices(centers, n_nodes: int, n_edges_per_node: int):
    # computes tensors of indices useful to convert between edge
    # and NEF layouts; the usage and function of nef_indices and
    # nef_to_edges_neighbor is clear in the edge_array_to_nef and
    # nef_array_to_edges functions below.
    # In particular:
    # nef_array = edge_array[nef_indices]
    # edge_array = nef_array[centers, nef_to_edges_neighbor]
    # The third output, nef_mask, is a mask that can be used to
    # filter out the padding values in the NEF array, as different
    # nodes will have, in general, different number of edges.

    bincount = torch.bincount(centers, minlength=n_nodes)

    arange = torch.arange(n_edges_per_node, device=centers.device)
    arange_expanded = arange.view(1, -1).expand(n_nodes, -1)
    nef_mask = arange_expanded < bincount.view(-1, 1)

    argsort = torch.argsort(centers, stable=True)

    nef_indices = torch.zeros(
        (n_nodes, n_edges_per_node), dtype=torch.long, device=centers.device
    )
    nef_indices[nef_mask] = argsort

    nef_to_edges_neighbor = torch.empty_like(centers, dtype=torch.long)
    nef_to_edges_neighbor[argsort] = arange_expanded[nef_mask]

    return nef_indices, nef_to_edges_neighbor, nef_mask


def get_corresponding_edges(array):
    # computes the corresponding edge (i.e., the edge that goes in the
    # opposite direction) for each edge in the array; this is useful
    # in the message-passing operation

    if array.numel() == 0:
        return torch.empty((0,), dtype=array.dtype, device=array.device)

    array = array.to(torch.int64)  # avoid overflow

    centers = array[:, 0]
    neighbors = array[:, 1]
    cell_shifts_x = array[:, 2]
    cell_shifts_y = array[:, 3]
    cell_shifts_z = array[:, 4]

    # will be useful later
    negative_cell_shifts_x = -cell_shifts_x
    negative_cell_shifts_y = -cell_shifts_y
    negative_cell_shifts_z = -cell_shifts_z

    # create a unique identifier for each edge
    # first, we shift the cell_shifts so that the minimum value is 0
    min_cell_shift_x = cell_shifts_x.min()
    cell_shifts_x = cell_shifts_x - min_cell_shift_x
    negative_cell_shifts_x = negative_cell_shifts_x - min_cell_shift_x

    min_cell_shift_y = cell_shifts_y.min()
    cell_shifts_y = cell_shifts_y - min_cell_shift_y
    negative_cell_shifts_y = negative_cell_shifts_y - min_cell_shift_y

    min_cell_shift_z = cell_shifts_z.min()
    cell_shifts_z = cell_shifts_z - min_cell_shift_z
    negative_cell_shifts_z = negative_cell_shifts_z - min_cell_shift_z

    max_centers_neigbors = centers.max() + 1  # same as neighbors.max() + 1
    max_shift_x = cell_shifts_x.max() + 1
    max_shift_y = cell_shifts_y.max() + 1
    max_shift_z = cell_shifts_z.max() + 1

    size_1 = max_shift_z
    size_2 = max_shift_y * size_1
    size_3 = max_shift_x * size_2
    size_4 = max_centers_neigbors * size_3

    unique_id = (
        centers * size_4
        + neighbors * size_3
        + cell_shifts_x * size_2
        + cell_shifts_y * size_1
        + cell_shifts_z
    )

    # the inverse is the same, but centers and neighbors are swapped
    # and we use the negative values of the cell_shifts
    unique_id_inverse = (
        neighbors * size_4
        + centers * size_3
        + negative_cell_shifts_x * size_2
        + negative_cell_shifts_y * size_1
        + negative_cell_shifts_z
    )

    unique_id_argsort = unique_id.argsort()
    unique_id_inverse_argsort = unique_id_inverse.argsort()

    corresponding_edges = torch.empty_like(centers)
    corresponding_edges[unique_id_argsort] = unique_id_inverse_argsort

    return corresponding_edges.to(array.dtype)


def edge_array_to_nef(
    edge_array,
    nef_indices,
    mask: Optional[torch.Tensor] = None,
    fill_value: float = 0.0,
):
    # converts an edge array to a NEF array
    if mask is None:
        return edge_array[nef_indices]
    else:
        return torch.where(
            mask.reshape(mask.shape + (1,) * (len(edge_array.shape) - 1)),
            edge_array[nef_indices],
            fill_value,
        )


def nef_array_to_edges(nef_array, centers, nef_to_edges_neighbor):
    # converts a NEF array to an edge array. Most often, this converts
    # a NEF array (three-dimensional, where the dimensions are the nodes,
    # the edges, and the features) to an edge array (two-dimensional,
    # where the dimensions are the edges and the features).
    return nef_array[centers, nef_to_edges_neighbor]



from typing import List

import numpy as np
import torch


class CombinedDataLoader:
    """
    Combines multiple dataloaders into a single dataloader.

    This is useful for learning from multiple datasets at the same time,
    each of which may have different batch sizes, properties, etc.

    :param dataloaders: list of dataloaders to combine
    :param shuffle: whether to shuffle the combined dataloader (this does not
        act on the individual batches, but it shuffles the order in which
        they are returned)

    :return: the combined dataloader
    """

    def __init__(self, dataloaders: List[torch.utils.data.DataLoader], shuffle: bool):
        self.dataloaders = dataloaders
        self.shuffle = shuffle

        # Create the indices:
        self.indices = list(range(len(self)))

        # Shuffle the indices if requested
        if self.shuffle:
            np.random.shuffle(self.indices)

        self.reset()

    def reset(self):
        self.current_index = 0
        self.full_list = [batch for dl in self.dataloaders for batch in dl]

    def __iter__(self):
        return self

    def __next__(self):
        if self.current_index >= len(self.indices):
            self.reset()  # Reset the index for the next iteration
            raise StopIteration

        idx = self.indices[self.current_index]
        self.current_index += 1
        return self.full_list[idx]

    def __len__(self):
        """Total number of batches in all dataloaders.

        This returns the total number of batches in all dataloaders
        (as opposed to the total number of samples or the number of
        individual dataloaders).

        :return: the total number of batches in all dataloaders
        """
        return sum(len(dl) for dl in self.dataloaders)



from typing import Dict, List

import torch
from metatensor.torch import TensorBlock, TensorMap
from metatensor.torch.atomistic import System


def average_by_num_atoms(
    tensor_map_dict: Dict[str, TensorMap],
    systems: List[System],
    per_structure_keys: List[str],
) -> Dict[str, TensorMap]:
    """
    Averages a dictionary of ``TensorMap`` objects by the number of
    atoms in each system.

    This function averages by the number of atoms in each system. Targets that
    are present in ``per_structure_keys`` will not be averaged.

    :param tensor_map_dict: A dictionary of ``TensorMap`` objects.
    :param systems: The systems used to compute the predictions.
    :param per_structure_keys: A list of keys whose corresponding
        ``TensorMap`` objects that should not be averaged.

    :return: The dictionary of averaged ``TensorMap`` objects.
    """
    averaged_tensor_map_dict = {}
    device = systems[0].device
    num_atoms = torch.tensor([len(s) for s in systems], device=device)
    for key in tensor_map_dict.keys():
        if key in per_structure_keys:
            averaged_tensor_map_dict[key] = tensor_map_dict[key]
        else:
            averaged_tensor_map_dict[key] = divide_by_num_atoms(
                tensor_map_dict[key], num_atoms
            )
    return averaged_tensor_map_dict


def divide_by_num_atoms(tensor_map: TensorMap, num_atoms: torch.Tensor) -> TensorMap:
    """Takes the average values per atom of a ``TensorMap``.

    Since some quantities might already be per atom (e.g., atomic energies
    or position gradients), this function only divides a block (or gradient
    block) by the number of atoms if the block's samples do not contain
    the "atom" key. In practice, this guarantees the desired behavior for
    the majority of the cases, including energies, forces, and virials, where
    the energies and virials should be divided by the number of atoms, while
    the forces should not.

    :param tensor_map: The input tensor map.
    :param num_atoms: The number of atoms in each system.

    :return: A new tensor map with the values divided by the number of atoms.
    """

    blocks = []
    for block in tensor_map.blocks():
        if "atom" in block.samples.names:
            new_block = block
        else:
            values = block.values / num_atoms.view(
                -1, *[1] * (len(block.values.shape) - 1)
            )
            new_block = TensorBlock(
                values=values,
                samples=block.samples,
                components=block.components,
                properties=block.properties,
            )
            for gradient_name, gradient in block.gradients():
                if "atom" in gradient.samples.names:
                    new_gradient = gradient
                else:
                    values = gradient.values / num_atoms.view(
                        -1, *[1] * (len(gradient.values.shape) - 1)
                    )
                    new_gradient = TensorBlock(
                        values=values,
                        samples=gradient.samples,
                        components=gradient.components,
                        properties=gradient.properties,
                    )
                new_block.add_gradient(gradient_name, new_gradient)
        blocks.append(new_block)

    return TensorMap(
        keys=tensor_map.keys,
        blocks=blocks,
    )



from typing import Dict, List, Optional, Tuple

import pet_neighbors_convert  # noqa: F401
import torch
from metatensor.torch import Labels
from metatensor.torch.atomistic import NeighborListOptions, System


def collate_graph_dicts(
    graph_dicts: List[Dict[str, torch.Tensor]],
) -> Dict[str, torch.Tensor]:
    """
    Collates a list of graphs into a single graph.

    :param graph_dicts: A list of graphs to be collated.

    :return: The collated grap (batch).
    """
    device = graph_dicts[0]["x"].device
    simple_concatenate_keys: List[str] = [
        "central_species",
        "x",
        "neighbor_species",
        "neighbors_pos",
        "nums",
        "mask",
    ]
    cumulative_adjust_keys: List[str] = ["neighbors_index"]

    result: Dict[str, List[torch.Tensor]] = {}

    n_nodes_cumulative: int = 0

    number_of_graphs: int = int(len(graph_dicts))

    for index in range(number_of_graphs):
        graph: Dict[str, torch.Tensor] = graph_dicts[index]

        for key in simple_concatenate_keys:
            if key not in result:
                result[key] = [graph[key]]
            else:
                result[key].append(graph[key])

        for key in cumulative_adjust_keys:
            if key not in result:
                graph_key: torch.Tensor = graph[key]

                now: List[torch.Tensor] = [graph_key + n_nodes_cumulative]
                result[key] = now

            else:
                graph_key_2: torch.Tensor = graph[key]

                now_2: torch.Tensor = graph_key_2 + n_nodes_cumulative
                result[key].append(now_2)

        n_atoms: int = graph["central_species"].shape[0]

        index_repeated: torch.Tensor = torch.LongTensor([index for _ in range(n_atoms)])
        if "batch" not in result.keys():
            result["batch"] = [index_repeated]
        else:
            result["batch"].append(index_repeated)

        n_nodes_cumulative += n_atoms

    result_final: Dict[str, torch.Tensor] = {}
    for key in simple_concatenate_keys + cumulative_adjust_keys:
        now_3: List[torch.Tensor] = []
        for el in result[key]:
            now_3.append(el)

        result_final[key] = torch.cat(now_3, dim=0)

    result_final["batch"] = torch.cat(result["batch"], dim=0).to(device)
    return result_final


def get_max_num_neighbors(systems: List[System], options: NeighborListOptions):
    """
    Calculates the maximum number of neighbors that atoms in a list of systems have.

    """
    max_system_num_neighbors = []
    for system in systems:
        nl = system.get_neighbor_list(options)
        i_list = nl.samples.column("first_atom")
        if len(i_list) == 0:
            max_atom_num_neighbors = torch.tensor(
                0, device=i_list.device, dtype=i_list.dtype
            )
        else:
            max_atom_num_neighbors = torch.bincount(i_list).max()
        max_system_num_neighbors.append(max_atom_num_neighbors)
    return torch.stack(max_system_num_neighbors).max()


def get_central_species(
    system: System, all_species: torch.Tensor, unique_index: torch.Tensor
) -> torch.Tensor:
    """
    Returns the indices of the species of the central atoms in the system
    in a list of all species.

    """
    species = system.types[unique_index]
    tmp_index_1, tmp_index_2 = torch.where(all_species.unsqueeze(1) == species)
    index = torch.argsort(tmp_index_2)
    return tmp_index_1[index]


def write_system_data(
    system: System,
    options: NeighborListOptions,
    selected_atoms_index: torch.Tensor,
):
    nl = system.get_neighbor_list(options)
    i_list = nl.samples.column("first_atom")
    j_list = nl.samples.column("second_atom")
    S_list = torch.cat(
        (
            nl.samples.column("cell_shift_a")[None],
            nl.samples.column("cell_shift_b")[None],
            nl.samples.column("cell_shift_c")[None],
        )
    ).transpose(0, 1)
    D_list = nl.values[:, :, 0]
    positions = system.positions
    types = system.types
    cell = system.cell
    torch.save(
        {
            "i_list": i_list,
            "j_list": j_list,
            "S_list": S_list,
            "D_list": D_list,
            "positions": positions,
            "types": types,
            "cell": cell,
            "selected_atoms_index": selected_atoms_index,
        },
        "system_data.pt",
    )


def write_batch_dict(batch_dict: Dict[str, torch.Tensor]):
    torch.save(batch_dict, "batch_dict.pt")


def remap_to_contiguous_indexing(
    i_list: torch.Tensor,
    j_list: torch.Tensor,
    unique_neighbors_index: torch.Tensor,
    unique_index: torch.Tensor,
    device: torch.device,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    This helper function remaps the indices of center and neighbor atoms
    from arbitrary indexing to contgious indexing, i.e.

    from
    0, 1, 2, 54, 55, 56
    to
    0, 1, 2, 3, 4, 5.

    This remapping is required by internal implementation of PET neighbor lists, where
    indices of the atoms cannot exceed the total amount of atoms in the system.

    Shifted indices come from LAMMPS neighborlists in the case of domain decomposition
    enabled, since they contain not only the atoms in the unit cell, but also so-called
    ghost atoms, which may have a different indexing. Thus, to avoid further errors, we
    remap the indices to a contiguous format.

    """
    index_map = torch.empty(
        int(unique_index.max().item()) + 1, dtype=torch.int64, device=device
    )
    index_map[unique_index] = torch.arange(len(unique_index), device=device)
    i_list = index_map[i_list]
    j_list = index_map[j_list]
    unique_neighbors_index = index_map[unique_neighbors_index]
    return i_list, j_list, unique_neighbors_index


def get_system_batch_dict(
    system: System,
    options: NeighborListOptions,
    all_species: torch.Tensor,
    max_num_neighbors: torch.Tensor,
    selected_atoms_index: torch.Tensor,
    device: torch.device,
    debug: bool = False,
) -> Dict[str, torch.Tensor]:
    if debug:
        write_system_data(system, options, selected_atoms_index)
    nl = system.get_neighbor_list(options)
    i_list = nl.samples.column("first_atom")
    j_list = nl.samples.column("second_atom")

    unique_neighbors_index, _ = torch.unique(i_list, return_counts=True)
    unique_index = torch.unique(
        torch.cat((selected_atoms_index, unique_neighbors_index))
    )

    # We calculate the actual size of the system, which is the number of
    # unique atoms in the system.
    # This is required for LAMMPS interface, because by default
    # it produces the system with both local and ghost atoms.
    actual_system_size = len(unique_index)

    # Then we remap the indices of the atoms to a contiguous format.
    # Also see the docstring of the function for more details.
    i_list, j_list, unique_neighbors_index = remap_to_contiguous_indexing(
        i_list, j_list, unique_neighbors_index, unique_index, device
    )

    # We sort the indices of the atoms in the system, to join the
    # periodic images of the same atom together. Otherwise, the
    # neighbor list may have a discontinuous indexing, like:
    # >>> i_list
    # tensor([0, 1, 2, 0, 1, 2])
    # instead of
    # >>> i_list
    # tensor([0, 0, 1, 1, 2, 2])
    # and we heavily rely on the fact that the indices of the atoms
    # are contiguous below.
    index = torch.argsort(i_list, stable=True)
    j_list = j_list[index].contiguous()
    i_list = i_list[index].contiguous()
    S_list: torch.Tensor = (
        torch.cat(
            (
                nl.samples.column("cell_shift_a")[None],
                nl.samples.column("cell_shift_b")[None],
                nl.samples.column("cell_shift_c")[None],
            )
        )
        .transpose(0, 1)[index]
        .to(torch.int64)
    ).contiguous()
    D_list = (
        system.positions[j_list]
        - system.positions[i_list]
        + S_list.to(system.cell.dtype) @ system.cell
    ).contiguous()

    species = system.types[unique_index].to(torch.int64)

    res = torch.ops.neighbors_convert.process(
        i_list,
        j_list,
        S_list,
        D_list,
        max_num_neighbors,
        torch.tensor(actual_system_size),
        species,
        all_species,
    )

    # `neighbors_index`: indices of the neighbors
    # from j_list for each central atom
    neighbors_index = res[0]
    # `relative_positions`: displacements of the neighbors
    # from j_list relative to each central atom
    relative_positions = res[1]
    # `nums`: number of neighbors for each central atom
    nums = res[2]
    # `mask`: for each central atom extracts the actual
    # neighbors data from the remaining tensors (i.e neighbors_index,
    # relative_positions, etc). The reason for this, is a padding that
    # enlarges the leghts of these tensors for each central atom up to
    # maximum number of neighbors across all the central atoms, to fix
    # the dimensionality of the neighborlist and fit is into a single
    # tensor. Thus, mask essentially show, where the padding is used, and
    # where is the actual NL data.
    mask = res[3]
    # `neighbor_species`: the indices of the species of the neighbors in the
    # all_species tensor.
    neighbor_species = res[4]
    # `neighbors_pos`: the reversed neighbors index, which is used in the
    # PET model to account for edge information update not only with the central
    # atoms data, but also with the neighbors data. This requires knowing the
    # reversed indices of the neighbors in the neighbor list.
    #
    # The reversed neighbor index is basically the index of the central atom in the
    # neighbor list of the neighbor atom.
    #
    # Example:
    # >>> neighbors_index
    # tensor([[25, 28, 39, ...],
    #         ...
    #         [ 2,  3,  4, ...]])
    # >>> neighbors_pos # reversed_neighbors_index
    # tensor([[ 3,  4,  1, ...],
    #         ...
    #         [ 3,  8,  7, ...]])
    #
    # The first atom has the neighbors with indices 25, 28, 39, etc.,
    # That means, in the list of neighbors of 25th atom, 4th atom will
    # have the index 0 (i.e. that will be the first atom).
    #
    # >>> neighbors_index[25]
    # tensor([45, 29, 47,  0, ...])
    # >>> neighbors_index[25][3]
    # tensor(0)
    #
    # and this is the element [0, 0] of the `neighbors_pos`.
    #
    # We also demand the reversed cell shift vector to be the opposite
    # of the original cell shift vector. This is because sometimes the
    # central atom may have two neighbors, which are the same atom, but
    # different periodic images.
    neighbors_pos = res[5]
    # central_species: the indices of species of the central
    # atoms of the system in the all_species tensor.
    central_species = res[6]

    system_dict = {
        "central_species": central_species,
        "neighbor_species": neighbor_species,
        "x": relative_positions,
        "neighbors_pos": neighbors_pos,
        "neighbors_index": neighbors_index,
        "nums": nums,
        "mask": mask,
    }
    if debug:
        write_batch_dict(system_dict)
    return system_dict


def systems_to_batch_dict(
    systems: List[System],
    options: NeighborListOptions,
    all_species_list: List[int],
    selected_atoms: Optional[Labels] = None,
) -> Dict[str, torch.Tensor]:
    """
    Converts a standard input data format of `metatrain` to a
    PyTorch Geometric `Batch` object, compatible with `PET` model.

    :param systems: The list of systems in `metatensor.torch.atomistic.System`
    format, that needs to be converted.
    :param options: A `NeighborListOptions` objects specifying the parameters
    for a neighbor list, which will be used during the convertation.
    :param all_species: A `torch.Tensor` with all the species present in the
    systems.

    :return: Batch compatible with PET.
    """
    device = systems[0].positions.device
    all_species = torch.tensor(all_species_list, device=device)
    batch: List[Dict[str, torch.Tensor]] = []
    max_num_neighbors = get_max_num_neighbors(systems, options)
    for i, system in enumerate(systems):
        if selected_atoms is not None:
            selected_atoms_index = selected_atoms.values[:, 1][
                selected_atoms.values[:, 0] == i
            ]
        else:
            selected_atoms_index = torch.arange(len(system), device=device)
        system_dict = get_system_batch_dict(
            system,
            options,
            all_species,
            max_num_neighbors,
            selected_atoms_index,
            device,
        )
        batch.append(system_dict)
    return collate_graph_dicts(batch)



from typing import Tuple


def get_gradient_units(base_unit: str, gradient_name: str, length_unit: str) -> str:
    """
    Get the gradient units based on the unit of the base quantity.

    For example, if the base unit is "<unit>" and the gradient name is
    "positions", the gradient unit will be "<unit>/<length_unit>".

    :param base_unit: The unit of the base quantity.
    :param gradient_name: The name of the gradient.
    :param length_unit: The unit of lengths.

    :return: The unit of the gradient.
    """
    if base_unit == "":
        return ""  # unknown unit for base quantity -> unknown unit for gradient
    if length_unit == "angstrom":
        length_unit = "Å"  # prettier
    if gradient_name == "positions":
        return base_unit + "/" + length_unit
    elif gradient_name == "strain":
        return base_unit  # strain is dimensionless
    else:
        raise ValueError(f"Unknown gradient name: {gradient_name}")


def ev_to_mev(value: float, unit: str) -> Tuple[float, str]:
    """
    If the `unit` starts with eV, converts the `value` and its
    corresponding `unit` to meV. Otherwise, returns the input.

    :param value: The value (potentially in eV or a derived quantity of eV).
    :param unit: The unit of the value.

    :return: If the `value` is in meV (or a derived quantity), the value and
        the corresponding unit where eV is converted to meV. Otherwise, the input.
    """
    if unit.startswith("eV") or unit.startswith("ev"):
        return value * 1000.0, (
            unit.replace("eV", "meV")
            if unit.startswith("eV")
            else unit.replace("ev", "mev")
        )
    else:
        return value, unit



import re
import warnings
from pathlib import Path
from typing import Any, Optional, Union
from urllib.parse import unquote, urlparse
from urllib.request import urlretrieve

import torch
from metatensor.torch.atomistic import check_atomistic_model, load_atomistic_model

from ..utils.architectures import find_all_architectures
from .architectures import import_architecture


def check_file_extension(
    filename: Union[str, Path], extension: str
) -> Union[str, Path]:
    """Check the file extension of a file name and adds if it is not present.

    If ``filename`` does not end with ``extension`` the ``extension`` is added and a
    warning will be issued.

    :param filename: Name of the file to be checked.
    :param extension: Expected file extension i.e. ``.txt``.
    :returns: Checked and probably extended file name.
    """
    path_filename = Path(filename)

    if path_filename.suffix != extension:
        warnings.warn(
            f"The file name should have a '{extension}' file extension. The user "
            f"requested the file with name '{filename}', but it will be saved as "
            f"'{filename}{extension}'.",
            stacklevel=1,
        )
        path_filename = path_filename.parent / (path_filename.name + extension)

    if type(filename) is str:
        return str(path_filename)
    else:
        return path_filename


def is_exported_file(path: str) -> bool:
    """
    Check if a saved model file has been exported to a ``MetatensorAtomisticModel``.

    The functions uses :py:func:`metatensor.torch.atomistic.check_atomistic_model` to
    verify.

    :param path: model path
    :return: :py:obj:`True` if the ``model`` has been exported, :py:obj:`False`
        otherwise.

    .. seealso::

        :py:func:`metatensor.torch.atomistic.is_atomistic_model` to verify if an already
        loaded model is exported.
    """
    try:
        check_atomistic_model(str(path))
        return True
    except ValueError:
        return False


def _hf_hub_download_url(url: str, token: Optional[str] = None) -> str:
    """Wrapper around `hf_hub_download` allowing passing the URL directly.

    Function is in inverse of `hf_hub_url`
    """
    try:
        from huggingface_hub import hf_hub_download
    except ImportError:
        raise ImportError(
            "To download a private model please install the `huggingface_hub` package "
            "with pip (`pip install huggingface_hub`)."
        )

    pattern = re.compile(
        r"(?P<endpoint>https://[^/]+)/"
        r"(?P<repo_id>[^/]+/[^/]+)/"
        r"resolve/"
        r"(?P<revision>[^/]+)/"
        r"(?P<filename>.+)"
    )

    match = pattern.match(url)

    if not match:
        raise ValueError(f"URL '{url}' has an invalid format for the Hugging Face Hub.")

    endpoint = match.group("endpoint")
    repo_id = match.group("repo_id")
    revision = unquote(match.group("revision"))
    filename = unquote(match.group("filename"))

    # Extract subfolder if applicable
    parts = filename.split("/", 1)
    if len(parts) == 2:
        subfolder, filename = parts
    else:
        subfolder = None

    return hf_hub_download(
        repo_id=repo_id,
        filename=filename,
        subfolder=subfolder,
        repo_type=None,
        revision=revision,
        endpoint=endpoint,
        token=token,
    )


def load_model(
    path: Union[str, Path],
    extensions_directory: Optional[Union[str, Path]] = None,
    token: Optional[str] = None,
) -> Any:
    """Load checkpoints and exported models from an URL or a local file.

    If an exported model should be loaded and requires compiled extensions, their
    location should be passed using the ``extensions_directory`` parameter.

    After reading a checkpoint, the returned model can be exported with the model's own
    ``export()`` method.

    :param path: local or remote path to a model. For supported URL schemes see
        :py:class:`urllib.request`
    :param extensions_directory: path to a directory containing all extensions required
        by an *exported* model
    :param token: HuggingFace API token to download (private) models from HuggingFace

    :raises ValueError: if ``path`` is a YAML option file and no model
    :raises ValueError: if no ``archietcture_name`` is found in the checkpoint
    :raises ValueError: if the ``architecture_name`` is not found in the available
        architectures
    """

    if Path(path).suffix in [".yaml", ".yml"]:
        raise ValueError(
            f"path '{path}' seems to be a YAML option file and not a model"
        )

    path = str(path)

    # Download remote model
    # TODO(@PicoCentauri): Introduce caching for remote models
    if urlparse(path).scheme:
        if token is None:
            path, _ = urlretrieve(path)
        else:
            path = _hf_hub_download_url(path, token=token)

    if is_exported_file(path):
        return load_atomistic_model(path, extensions_directory=extensions_directory)
    else:  # model is a checkpoint
        checkpoint = torch.load(path, weights_only=False, map_location="cpu")

        architecture_name = checkpoint["architecture_name"]
        if architecture_name not in find_all_architectures():
            raise ValueError(
                f"Checkpoint architecture '{architecture_name}' not found "
                "in the available architectures. Available architectures are: "
                f"{find_all_architectures()}"
            )
        architecture = import_architecture(architecture_name)

        try:
            return architecture.__model__.load_checkpoint(path)
        except Exception as err:
            raise ValueError(
                f"path '{path}' is not a valid checkpoint for the {architecture_name} "
                "architecture"
            ) from err



import argparse


class CustomHelpFormatter(argparse.RawDescriptionHelpFormatter):
    """Descriptions formatter showing positional arguments before optionals."""

    def _format_usage(self, usage, actions, groups, prefix):
        if usage is None:
            # split optionals from positionals
            optionals = []
            positionals = []
            for action in actions:
                if action.option_strings:
                    optionals.append(action)
                else:
                    positionals.append(action)

            prog = "%(prog)s" % dict(prog=self._prog)

            # build full usage string
            format = self._format_actions_usage
            action_usage = format(positionals + optionals, groups)
            usage = " ".join([s for s in [prog, action_usage] if s])

        # Call the superclass method to format the usage
        return super()._format_usage(usage, actions, groups, prefix)



import copy
from typing import Dict, List, Tuple

import torch.distributed
from metatensor.torch import TensorMap


class RMSEAccumulator:
    """Accumulates the RMSE between predictions and targets for an arbitrary
    number of keys, each corresponding to one target.

    :param separate_blocks: if true, the RMSE will be computed separately for each
        block in the target and prediction ``TensorMap`` objects.
    """

    def __init__(self, separate_blocks: bool = False) -> None:
        """Initialize the accumulator."""
        self.information: Dict[str, Tuple[float, int]] = {}
        self.separate_blocks = separate_blocks

    def update(self, predictions: Dict[str, TensorMap], targets: Dict[str, TensorMap]):
        """Updates the accumulator with new predictions and targets.

        :param predictions: A dictionary of predictions, where the keys correspond
            to the keys in the targets dictionary, and the values are the predictions.

        :param targets: A dictionary of targets, where the keys correspond to the keys
            in the predictions dictionary, and the values are the targets.
        """

        for key, target in targets.items():
            prediction = predictions[key]
            for block_key in target.keys:
                target_block = target.block(block_key)
                prediction_block = prediction.block(block_key)

                key_to_write = copy.deepcopy(key)
                if self.separate_blocks:
                    key_to_write += "("
                    for name, value in zip(block_key.names, block_key.values):
                        key_to_write += f"{name}={int(value)},"
                    key_to_write = key_to_write[:-1]
                    key_to_write += ")"

                if key_to_write not in self.information:  # create key if not present
                    self.information[key_to_write] = (0.0, 0)

                self.information[key_to_write] = (
                    self.information[key_to_write][0]
                    + ((prediction_block.values - target_block.values) ** 2)
                    .sum()
                    .item(),
                    self.information[key_to_write][1] + prediction_block.values.numel(),
                )

                for gradient_name, target_gradient in target_block.gradients():
                    if (
                        f"{key_to_write}_{gradient_name}_gradients"
                        not in self.information
                    ):
                        self.information[
                            f"{key_to_write}_{gradient_name}_gradients"
                        ] = (0.0, 0)
                    prediction_gradient = prediction_block.gradient(gradient_name)
                    self.information[f"{key_to_write}_{gradient_name}_gradients"] = (
                        self.information[f"{key_to_write}_{gradient_name}_gradients"][0]
                        + ((prediction_gradient.values - target_gradient.values) ** 2)
                        .sum()
                        .item(),
                        self.information[f"{key_to_write}_{gradient_name}_gradients"][1]
                        + prediction_gradient.values.numel(),
                    )

    def finalize(
        self,
        not_per_atom: List[str],
        is_distributed: bool = False,
        device: torch.device = None,
    ) -> Dict[str, float]:
        """Finalizes the accumulator and returns the RMSE for each key.

        All keys will be returned as "{key} RMSE (per atom)" in the output dictionary,
        unless ``key`` contains one or more of the strings in ``not_per_atom``,
        in which case "{key} RMSE" will be returned.

        :param not_per_atom: a list of strings. If any of these strings are present in
            a key, the RMSE key will not be labeled as "(per atom)".
        :param is_distributed: if true, the RMSE will be computed across all ranks
            of the distributed system.
        :param device: the local device to use for the computation. Only needed if
            ``is_distributed`` is :obj:`python:True`.
        """

        if is_distributed:
            for key, value in self.information.items():
                sse = torch.tensor(value[0]).to(device)
                n_elems = torch.tensor(value[1]).to(device)
                torch.distributed.all_reduce(sse)
                torch.distributed.all_reduce(n_elems)
                self.information[key] = (sse.item(), n_elems.item())  # type: ignore

        finalized_info = {}
        for key, value in self.information.items():
            if any([s in key for s in not_per_atom]):
                out_key = f"{key} RMSE"
            else:
                out_key = f"{key} RMSE (per atom)"
            finalized_info[out_key] = (value[0] / value[1]) ** 0.5

        return finalized_info


class MAEAccumulator:
    """Accumulates the MAE between predictions and targets for an arbitrary
    number of keys, each corresponding to one target.

    :param separate_blocks: if true, the RMSE will be computed separately for each
        block in the target and prediction ``TensorMap`` objects.
    """

    def __init__(self, separate_blocks: bool = False) -> None:
        """Initialize the accumulator."""
        self.information: Dict[str, Tuple[float, int]] = {}
        self.separate_blocks = separate_blocks

    def update(self, predictions: Dict[str, TensorMap], targets: Dict[str, TensorMap]):
        """Updates the accumulator with new predictions and targets.

        :param predictions: A dictionary of predictions, where the keys correspond
            to the keys in the targets dictionary, and the values are the predictions.

        :param targets: A dictionary of targets, where the keys correspond to the keys
            in the predictions dictionary, and the values are the targets.
        """

        for key, target in targets.items():
            prediction = predictions[key]
            for block_key in target.keys:
                target_block = target.block(block_key)
                prediction_block = prediction.block(block_key)

                key_to_write = copy.deepcopy(key)
                if self.separate_blocks:
                    key_to_write += "("
                    for name, value in zip(block_key.names, block_key.values):
                        key_to_write += f"{name}={int(value)},"
                    key_to_write = key_to_write[:-1]
                    key_to_write += ")"

                if key_to_write not in self.information:  # create key if not present
                    self.information[key_to_write] = (0.0, 0)

                self.information[key_to_write] = (
                    self.information[key_to_write][0]
                    + (prediction_block.values - target_block.values)
                    .abs()
                    .sum()
                    .item(),
                    self.information[key_to_write][1] + prediction_block.values.numel(),
                )

                for gradient_name, target_gradient in target_block.gradients():
                    if (
                        f"{key_to_write}_{gradient_name}_gradients"
                        not in self.information
                    ):
                        self.information[
                            f"{key_to_write}_{gradient_name}_gradients"
                        ] = (0.0, 0)
                    prediction_gradient = prediction_block.gradient(gradient_name)
                    self.information[f"{key_to_write}_{gradient_name}_gradients"] = (
                        self.information[f"{key_to_write}_{gradient_name}_gradients"][0]
                        + (prediction_gradient.values - target_gradient.values)
                        .abs()
                        .sum()
                        .item(),
                        self.information[f"{key_to_write}_{gradient_name}_gradients"][1]
                        + prediction_gradient.values.numel(),
                    )

    def finalize(
        self,
        not_per_atom: List[str],
        is_distributed: bool = False,
        device: torch.device = None,
    ) -> Dict[str, float]:
        """Finalizes the accumulator and returns the MAE for each key.

        All keys will be returned as "{key} MAE (per atom)" in the output dictionary,
        unless ``key`` contains one or more of the strings in ``not_per_atom``,
        in which case "{key} MAE" will be returned.

        :param not_per_atom: a list of strings. If any of these strings are present in
            a key, the MAE key will not be labeled as "(per atom)".
        :param is_distributed: if true, the MAE will be computed across all ranks
            of the distributed system.
        :param device: the local device to use for the computation. Only needed if
            ``is_distributed`` is :obj:`python:True`.
        """

        if is_distributed:
            for key, value in self.information.items():
                sae = torch.tensor(value[0]).to(device)
                n_elems = torch.tensor(value[1]).to(device)
                torch.distributed.all_reduce(sae)
                torch.distributed.all_reduce(n_elems)
                self.information[key] = (sae.item(), n_elems.item())  # type: ignore

        finalized_info = {}
        for key, value in self.information.items():
            if any([s in key for s in not_per_atom]):
                out_key = f"{key} MAE"
            else:
                out_key = f"{key} MAE (per atom)"
            finalized_info[out_key] = value[0] / value[1]

        return finalized_info


def get_selected_metric(metric_dict: Dict[str, float], selected_metric: str) -> float:
    """
    Selects and/or calculates a (user-)selected metric from a dictionary of metrics.

    This is useful when choosing the best model from a training run.

    :param metric_dict: A dictionary of metrics, where the keys are the names of the
        metrics and the values are the corresponding values.
    :param selected_metric: The metric to return. This can be one of the following:
        - "loss": return the loss value
        - "rmse_prod": return the product of all RMSEs
        - "mae_prod": return the product of all MAEs
    """
    if selected_metric == "loss":
        metric = metric_dict["loss"]
    elif selected_metric == "rmse_prod":
        metric = 1
        for key in metric_dict:
            if "RMSE" in key:
                metric *= metric_dict[key]
    elif selected_metric == "mae_prod":
        metric = 1
        for key in metric_dict:
            if "MAE" in key:
                metric *= metric_dict[key]
    else:
        raise ValueError(
            f"Selected metric {selected_metric} not recognized. "
            "Please select from 'loss', 'rmse_prod', or 'mae_prod'."
        )
    return metric



import copy

import metatensor.torch
import pytest
import torch
from jsonschema.exceptions import ValidationError
from metatensor.torch.atomistic import ModelOutput, System
from omegaconf import OmegaConf

from metatrain.soap_bpnn import SoapBpnn
from metatrain.utils.architectures import check_architecture_options
from metatrain.utils.data import DatasetInfo
from metatrain.utils.data.target_info import (
    get_energy_target_info,
    get_generic_target_info,
)

from . import DEFAULT_HYPERS, MODEL_HYPERS


def test_prediction_subset_elements():
    """Tests that the model can predict on a subset of the elements it was trained
    on."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )

    model = SoapBpnn(MODEL_HYPERS, dataset_info)

    system = System(
        types=torch.tensor([6, 6]),
        positions=torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]]),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    model(
        [system],
        {"energy": model.outputs["energy"]},
    )


def test_prediction_subset_atoms():
    """Tests that the model can predict on a subset
    of the atoms in a system."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )

    model = SoapBpnn(MODEL_HYPERS, dataset_info)

    # Since we don't yet support atomic predictions, we will test this by
    # predicting on a system with two monomers at a large distance

    system_monomer = System(
        types=torch.tensor([7, 8, 8]),
        positions=torch.tensor(
            [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 2.0]],
        ),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )

    energy_monomer = model(
        [system_monomer],
        {"energy": ModelOutput(per_atom=False)},
    )

    system_far_away_dimer = System(
        types=torch.tensor([7, 7, 8, 8, 8, 8]),
        positions=torch.tensor(
            [
                [0.0, 0.0, 0.0],
                [0.0, 50.0, 0.0],
                [0.0, 0.0, 1.0],
                [0.0, 0.0, 2.0],
                [0.0, 51.0, 0.0],
                [0.0, 42.0, 0.0],
            ],
        ),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )

    selection_labels = metatensor.torch.Labels(
        names=["system", "atom"],
        values=torch.tensor([[0, 0], [0, 2], [0, 3]]),
    )

    energy_dimer = model(
        [system_far_away_dimer],
        {"energy": ModelOutput(per_atom=False)},
    )

    energy_monomer_in_dimer = model(
        [system_far_away_dimer],
        {"energy": ModelOutput(per_atom=False)},
        selected_atoms=selection_labels,
    )

    assert not metatensor.torch.allclose(
        energy_monomer["energy"], energy_dimer["energy"]
    )

    assert metatensor.torch.allclose(
        energy_monomer["energy"], energy_monomer_in_dimer["energy"]
    )


def test_output_last_layer_features():
    """Tests that the model can output its last layer features."""
    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )

    model = SoapBpnn(MODEL_HYPERS, dataset_info)

    system = System(
        types=torch.tensor([6, 1, 8, 7]),
        positions=torch.tensor(
            [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 3.0]],
        ),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )

    # last-layer features per atom:
    ll_output_options = ModelOutput(
        quantity="",
        unit="unitless",
        per_atom=True,
    )
    outputs = model(
        [system],
        {
            "energy": model.outputs["energy"],
            "features": ll_output_options,
            "mtt::aux::energy_last_layer_features": ll_output_options,
        },
    )
    assert "energy" in outputs
    assert "features" in outputs
    assert "mtt::aux::energy_last_layer_features" in outputs

    features = outputs["features"].block()
    assert features.samples.names == [
        "system",
        "atom",
    ]
    assert features.values.shape == (
        4,
        128,
    )
    assert features.properties.names == [
        "properties",
    ]

    last_layer_features = outputs["mtt::aux::energy_last_layer_features"].block()
    assert last_layer_features.samples.names == [
        "system",
        "atom",
    ]
    assert last_layer_features.values.shape == (
        4,
        128,
    )
    assert last_layer_features.properties.names == [
        "properties",
    ]

    # last-layer features per system:
    ll_output_options = ModelOutput(
        quantity="",
        unit="unitless",
        per_atom=False,
    )
    outputs = model(
        [system],
        {
            "energy": model.outputs["energy"],
            "features": ll_output_options,
            "mtt::aux::energy_last_layer_features": ll_output_options,
        },
    )
    assert "energy" in outputs
    assert "features" in outputs
    assert "mtt::aux::energy_last_layer_features" in outputs

    features = outputs["features"].block()
    assert features.samples.names == [
        "system",
    ]
    assert features.values.shape == (
        1,
        128,
    )
    assert features.properties.names == [
        "properties",
    ]

    assert outputs["mtt::aux::energy_last_layer_features"].block().samples.names == [
        "system"
    ]
    assert outputs["mtt::aux::energy_last_layer_features"].block().values.shape == (
        1,
        128,
    )
    assert outputs["mtt::aux::energy_last_layer_features"].block().properties.names == [
        "properties",
    ]


def test_output_per_atom():
    """Tests that the model can output per-atom quantities."""
    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )

    model = SoapBpnn(MODEL_HYPERS, dataset_info)

    system = System(
        types=torch.tensor([6, 1, 8, 7]),
        positions=torch.tensor(
            [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 3.0]],
        ),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )

    outputs = model(
        [system],
        {"energy": model.outputs["energy"]},
    )

    assert outputs["energy"].block().samples.names == ["system", "atom"]
    assert outputs["energy"].block().values.shape == (4, 1)


def test_fixed_composition_weights():
    """Tests the correctness of the json schema for fixed_composition_weights"""

    hypers = copy.deepcopy(DEFAULT_HYPERS)
    hypers["training"]["fixed_composition_weights"] = {
        "energy": {
            1: 1.0,
            6: 0.0,
            7: 0.0,
            8: 0.0,
            9: 3000.0,
        }
    }
    hypers = OmegaConf.create(hypers)
    check_architecture_options(name="soap_bpnn", options=OmegaConf.to_container(hypers))


def test_fixed_composition_weights_error():
    """Test that only inputd of type Dict[str, Dict[int, float]] are allowed."""
    hypers = copy.deepcopy(DEFAULT_HYPERS)
    hypers["training"]["fixed_composition_weights"] = {"energy": {"H": 300.0}}
    hypers = OmegaConf.create(hypers)
    with pytest.raises(ValidationError, match=r"'H' does not match '\^\[0-9\]\+\$'"):
        check_architecture_options(
            name="soap_bpnn", options=OmegaConf.to_container(hypers)
        )


@pytest.mark.parametrize("per_atom", [True, False])
def test_vector_output(per_atom):
    """Tests that the model can predict a (spherical) vector output."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "forces": get_generic_target_info(
                {
                    "quantity": "forces",
                    "unit": "",
                    "type": {
                        "spherical": {"irreps": [{"o3_lambda": 1, "o3_sigma": 1}]}
                    },
                    "num_subtargets": 100,
                    "per_atom": per_atom,
                }
            )
        },
    )

    model = SoapBpnn(MODEL_HYPERS, dataset_info)

    system = System(
        types=torch.tensor([6, 6]),
        positions=torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]]),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    model(
        [system],
        {"force": model.outputs["forces"]},
    )


@pytest.mark.parametrize("per_atom", [True, False])
def test_spherical_outputs(per_atom):
    """Tests that the model can predict a spherical target with multiple blocks."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "spherical_target": get_generic_target_info(
                {
                    "quantity": "",
                    "unit": "",
                    "type": {
                        "spherical": {
                            "irreps": [
                                {"o3_lambda": 0, "o3_sigma": 1},
                                {"o3_lambda": 2, "o3_sigma": 1},
                            ]
                        }
                    },
                    "num_subtargets": 100,
                    "per_atom": per_atom,
                }
            )
        },
    )

    model = SoapBpnn(MODEL_HYPERS, dataset_info)

    system = System(
        types=torch.tensor([6, 6]),
        positions=torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]]),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    outputs = model(
        [system],
        {"spherical_target": model.outputs["spherical_target"]},
    )
    assert len(outputs["spherical_target"]) == 2


def test_soap_bpnn_single_atom():
    """Tests that the model predicts zero energies on a single atom."""
    # (note that no composition energies are supplied or calculated here)

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "energy": get_energy_target_info({"quantity": "energy", "unit": "eV"})
        },
    )
    hypers = copy.deepcopy(MODEL_HYPERS)
    hypers["soap"]["density"]["center_atom_weight"] = 0.0
    model = SoapBpnn(hypers, dataset_info)

    system = System(
        types=torch.tensor([6]),
        positions=torch.tensor([[0.0, 0.0, 1.0]]),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    outputs = {"energy": ModelOutput(per_atom=False)}
    energy = model([system], outputs)["energy"].block().values.item()
    assert energy == 0.0



import pytest


pytest.importorskip("pet_neighbors_convert")

from urllib.parse import urlparse
from urllib.request import urlretrieve

import metatensor.torch
import torch
from metatensor.torch.atomistic import ModelOutput

from metatrain.experimental.nativepet import NativePET
from metatrain.experimental.nativepet.modules.compatibility import (
    convert_checkpoint_from_legacy_pet,
)
from metatrain.experimental.nativepet.modules.structures import systems_to_batch
from metatrain.experimental.nativepet.modules.utilities import cutoff_func
from metatrain.pet import PET
from metatrain.pet.modules.hypers import Hypers
from metatrain.pet.modules.pet import PET as RawPET
from metatrain.pet.utils import (
    systems_to_batch_dict,
)
from metatrain.utils.architectures import get_default_hypers
from metatrain.utils.data import DatasetInfo, read_systems
from metatrain.utils.data.target_info import get_energy_target_info
from metatrain.utils.neighbor_lists import get_system_with_neighbor_lists
from metatrain.utils.output_gradient import compute_gradient

from . import DATASET_PATH, DATASET_WITH_FORCES_PATH


DEFAULT_PET_HYPERS = get_default_hypers("pet")
DEFAULT_NATIVEPET_HYPERS = get_default_hypers("experimental.nativepet")

DEFAULT_PET_HYPERS["model"]["R_CUT"] = DEFAULT_NATIVEPET_HYPERS["model"]["cutoff"]
DEFAULT_PET_HYPERS["model"]["BLEND_NEIGHBOR_SPECIES"] = True
DEFAULT_PET_HYPERS["model"]["TRANSFORMER_N_HEAD"] = DEFAULT_NATIVEPET_HYPERS["model"][
    "num_heads"
]
DEFAULT_PET_HYPERS["model"]["N_GNN_LAYERS"] = DEFAULT_NATIVEPET_HYPERS["model"][
    "num_gnn_layers"
]
DEFAULT_PET_HYPERS["model"]["N_TRANS_LAYERS"] = DEFAULT_NATIVEPET_HYPERS["model"][
    "num_attention_layers"
]


def set_embedding_weights(nativepet_state_dict, pet_state_dict):
    weight = nativepet_state_dict["embedding.weight"].detach().clone()
    weight.requires_grad = True
    pet_state_dict["pet.embedding.weight"] = weight
    return pet_state_dict


def set_gnn_weights(nativepet_state_dict, pet_state_dict):
    for key in nativepet_state_dict.keys():
        if "gnn_layers" in key:
            weight = nativepet_state_dict[key].detach().clone()
            weight.requires_grad = True
            pet_key = "pet." + key
            if "edge_embedder" in pet_key:
                pet_key = pet_key.replace("edge_embedder", "r_embedding")
            if "node_embedder" in pet_key:
                pet_key = pet_key.replace("node_embedder", "central_embedder")
            pet_state_dict[pet_key] = weight
    return pet_state_dict


def set_heads_weights(nativepet_state_dict, pet_state_dict):
    for key in nativepet_state_dict.keys():
        if "head" in key and "linear" not in key:
            pet_key = "pet." + key.replace("energy.", "")
            if "node" in key:
                pet_key = pet_key.replace("node_", "")
            if "edge" in key:
                pet_key = pet_key.replace("edge_", "bond_")
            pet_key = pet_key.replace("node_", "")
            split_key = pet_key.split(".")
            split_key.insert(3, "nn")
            pet_key = ".".join(split_key)
            weight = nativepet_state_dict[key].detach().clone()
            weight.requires_grad = True
            pet_state_dict[pet_key] = weight
    return pet_state_dict


def set_last_layers_weights(nativepet_state_dict, pet_state_dict):
    for key in nativepet_state_dict.keys():
        if "last_layer" in key:
            layer_num = key.split(".")[2]
            if "edge" in key:
                pet_key = f"pet.bond_heads.{layer_num}.nn.4.linear."
            else:
                pet_key = f"pet.heads.{layer_num}.nn.4.linear."
            if "weight" in key:
                pet_key += "weight"
            else:
                pet_key += "bias"
            weight = nativepet_state_dict[key].detach().clone()
            weight.requires_grad = True
            pet_state_dict[pet_key] = weight
    return pet_state_dict


def ensure_embeddings_weights_equality(nativepet_model, pet_model):
    torch.testing.assert_close(
        nativepet_model.embedding.weight, pet_model.pet.embedding.weight
    )


def ensure_gnn_layers_weights_equality(nativepet_model, pet_model):
    for i in range(len(nativepet_model.gnn_layers)):
        nativepet_gnn_layer = nativepet_model.gnn_layers[i]
        pet_gnn_layer = pet_model.pet.gnn_layers[i]
        # Testing if the r_embedding layers are the same
        torch.testing.assert_close(
            nativepet_gnn_layer.edge_embedder.weight, pet_gnn_layer.r_embedding.weight
        )
        torch.testing.assert_close(
            nativepet_gnn_layer.edge_embedder.bias, pet_gnn_layer.r_embedding.bias
        )
        # Testing if the compress layers are the same
        for j in range(len(nativepet_gnn_layer.compress)):
            if not hasattr(nativepet_gnn_layer.compress[j], "weight"):
                continue
            nativepet_compress_layer = nativepet_gnn_layer.compress[j]
            pet_compress_layer = pet_gnn_layer.compress[j]
            torch.testing.assert_close(
                nativepet_compress_layer.weight, pet_compress_layer.weight
            )
            torch.testing.assert_close(
                nativepet_compress_layer.bias, pet_compress_layer.bias
            )
        # Testing if the neighbor_embedder layers are the same
        if hasattr(nativepet_gnn_layer.neighbor_embedder, "weight"):
            torch.testing.assert_close(
                nativepet_gnn_layer.neighbor_embedder.weight,
                pet_gnn_layer.neighbor_embedder.weight,
            )
        # Testing if the transformer layers are the same
        for j in range(len(nativepet_gnn_layer.trans.layers)):
            nativepet_trans_layer = nativepet_gnn_layer.trans.layers[j]
            pet_trans_layer = pet_gnn_layer.trans.layers[j]
            for k in range(len(nativepet_trans_layer.mlp)):
                nativepet_mlp_layer = nativepet_trans_layer.mlp[k]
                pet_mlp_layer = pet_trans_layer.mlp[k]
                if not hasattr(nativepet_mlp_layer, "weight"):
                    continue
                # Testing if the mlp layers are the same
                torch.testing.assert_close(
                    nativepet_mlp_layer.weight, pet_mlp_layer.weight
                )
                torch.testing.assert_close(nativepet_mlp_layer.bias, pet_mlp_layer.bias)
            # Testing if the attention layers are the same
            torch.testing.assert_close(
                nativepet_trans_layer.attention.input_linear.weight,
                pet_trans_layer.attention.input_linear.weight,
            )
            torch.testing.assert_close(
                nativepet_trans_layer.attention.output_linear.weight,
                pet_trans_layer.attention.output_linear.weight,
            )
            torch.testing.assert_close(
                nativepet_trans_layer.attention.input_linear.bias,
                pet_trans_layer.attention.input_linear.bias,
            )
            torch.testing.assert_close(
                nativepet_trans_layer.attention.output_linear.bias,
                pet_trans_layer.attention.output_linear.bias,
            )
            # Testing if attention layernorms are the same
            torch.testing.assert_close(
                nativepet_trans_layer.norm_attention.weight,
                pet_trans_layer.norm_attention.weight,
            )
            torch.testing.assert_close(
                nativepet_trans_layer.norm_attention.bias,
                pet_trans_layer.norm_attention.bias,
            )
            # Testing if mlp layernorms are the same
            torch.testing.assert_close(
                nativepet_trans_layer.norm_mlp.weight,
                pet_trans_layer.norm_mlp.weight,
            )
            torch.testing.assert_close(
                nativepet_trans_layer.norm_mlp.bias,
                pet_trans_layer.norm_mlp.bias,
            )


def ensure_node_heads_weights_equality(nativepet_model, pet_model):
    for i in range(len(nativepet_model.node_heads["energy"])):
        nativepet_head = nativepet_model.node_heads["energy"][i]
        pet_head = pet_model.pet.heads[i]
        # Testing if the linear layers are the same
        for j in range(len(nativepet_head)):
            if not hasattr(nativepet_head[j], "weight"):
                continue
            nativepet_linear_layer = nativepet_head[j]
            pet_linear_layer = pet_head.nn[j]
            torch.testing.assert_close(
                nativepet_linear_layer.weight, pet_linear_layer.weight
            )
            torch.testing.assert_close(
                nativepet_linear_layer.bias, pet_linear_layer.bias
            )


def ensure_edge_heads_weights_equality(nativepet_model, pet_model):
    for i in range(len(nativepet_model.edge_heads["energy"])):
        nativepet_bond_head = nativepet_model.edge_heads["energy"][i]
        pet_bond_head = pet_model.pet.bond_heads[i]
        # Testing if the linear layers are the same
        for j in range(len(nativepet_bond_head)):
            if not hasattr(nativepet_bond_head[j], "weight"):
                continue
            nativepet_linear_layer = nativepet_bond_head[j]
            pet_linear_layer = pet_bond_head.nn[j]
            torch.testing.assert_close(
                nativepet_linear_layer.weight, pet_linear_layer.weight
            )
            torch.testing.assert_close(
                nativepet_linear_layer.bias, pet_linear_layer.bias
            )


def ensure_node_last_layers_weights_equality(nativepet_model, pet_model):
    for i in range(len(nativepet_model.node_last_layers["energy"])):
        nativepet_last_layer = nativepet_model.node_last_layers["energy"][i][
            "energy___0"
        ]
        pet_last_layer = pet_model.pet.heads[i].nn[-1].linear
        torch.testing.assert_close(nativepet_last_layer.weight, pet_last_layer.weight)
        torch.testing.assert_close(nativepet_last_layer.bias, pet_last_layer.bias)


def ensure_edge_last_layers_weights_equality(nativepet_model, pet_model):
    for i in range(len(nativepet_model.edge_last_layers["energy"])):
        nativepet_last_layer = nativepet_model.edge_last_layers["energy"][i][
            "energy___0"
        ]
        pet_last_layer = pet_model.pet.bond_heads[i].nn[-1].linear
        torch.testing.assert_close(nativepet_last_layer.weight, pet_last_layer.weight)
        torch.testing.assert_close(nativepet_last_layer.bias, pet_last_layer.bias)


def get_identical_pet_models():
    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "energy": get_energy_target_info({"quantity": "energy", "unit": "eV"})
        },
    )
    nativepet_model = NativePET(DEFAULT_NATIVEPET_HYPERS["model"], dataset_info)
    pet_model = PET(DEFAULT_PET_HYPERS["model"], dataset_info)
    raw_pet_hypers = Hypers(pet_model.hypers)
    raw_pet = RawPET(raw_pet_hypers, 0.0, len(pet_model.atomic_types))
    pet_model.set_trained_model(raw_pet)

    nativepet_state_dict = nativepet_model.state_dict()
    pet_state_dict = pet_model.state_dict()

    # Setting the embedding layers to be the same weights
    pet_state_dict = set_embedding_weights(nativepet_state_dict, pet_state_dict)
    # Setting the GNN layers to be the same weights
    pet_state_dict = set_gnn_weights(nativepet_state_dict, pet_state_dict)
    # Setting the heads to be the same weights
    pet_state_dict = set_heads_weights(nativepet_state_dict, pet_state_dict)
    # Setting the last layers to be the same weights
    pet_state_dict = set_last_layers_weights(nativepet_state_dict, pet_state_dict)

    # Loading the updated state dict
    pet_model.load_state_dict(pet_state_dict)

    # Testing if embedding layers are the same
    ensure_embeddings_weights_equality(nativepet_model, pet_model)
    # Testing if GNN layers are the same
    ensure_gnn_layers_weights_equality(nativepet_model, pet_model)
    # Testing if the heads are the same
    ensure_node_heads_weights_equality(nativepet_model, pet_model)
    # Testing if the bond heads are the same
    ensure_edge_heads_weights_equality(nativepet_model, pet_model)
    # Testing if the last layers are the same
    ensure_node_last_layers_weights_equality(nativepet_model, pet_model)
    # Testing if the bond last layers are the same
    ensure_edge_last_layers_weights_equality(nativepet_model, pet_model)

    return nativepet_model, pet_model


def get_test_environment():
    nativepet_model, pet_model = get_identical_pet_models()

    systems_1 = read_systems(DATASET_PATH)[:5]
    systems_2 = read_systems(DATASET_WITH_FORCES_PATH)[:5]
    systems = systems_1 + systems_2
    for system in systems:
        system.positions.requires_grad_(True)
        get_system_with_neighbor_lists(
            system, nativepet_model.requested_neighbor_lists()
        )
    systems = [system.to(torch.float32) for system in systems]
    return nativepet_model, pet_model, systems


def test_batch_dict_compatability():
    """Tests that the batch dictionaries created by the
    PET and NativePET models are the same"""
    nativepet_model, pet_model, systems = get_test_environment()

    nl_options = nativepet_model.requested_neighbor_lists()[0]

    system_indices, sample_labels = nativepet_model._get_system_indices_and_labels(
        systems, device="cpu"
    )

    (
        element_indices_nodes,
        element_indices_neighbors,
        edge_vectors,
        padding_mask,
        neighbors_index,
        num_neghbors,
        reversed_neighbor_list,
    ) = systems_to_batch(
        systems,
        nl_options,
        nativepet_model.atomic_types,
        system_indices,
        nativepet_model.species_to_species_index,
        None,
    )

    nativepet_batch_dict = {
        "central_species": element_indices_nodes,
        "neighbor_species": element_indices_neighbors,
        "x": edge_vectors,
        "mask": ~padding_mask,
        "neighbors_index": neighbors_index,
        "nums": num_neghbors,
        "batch": system_indices,
        "neighbors_pos": reversed_neighbor_list,
    }

    pet_batch_dict = systems_to_batch_dict(
        systems, nl_options, pet_model.atomic_types, None
    )

    for key in nativepet_batch_dict:
        if key in ["nums", "batch", "central_species", "mask"]:
            assert torch.allclose(nativepet_batch_dict[key], pet_batch_dict[key])
        else:
            nativepet_mask = nativepet_batch_dict["mask"] is False
            pet_mask = pet_batch_dict["mask"] is False
            assert torch.allclose(
                nativepet_batch_dict[key][nativepet_mask],
                pet_batch_dict[key][pet_mask],
            )


def test_embeddings_compatibility():
    """Tests that neighbor species embeddings are the same for
    the PET and NativePET models"""
    nativepet_model, pet_model, systems = get_test_environment()

    nl_options = nativepet_model.requested_neighbor_lists()[0]
    batch_dict = systems_to_batch_dict(
        systems, nl_options, nativepet_model.atomic_types, None
    )

    nativepet_embeddings = nativepet_model.embedding(batch_dict["neighbor_species"])
    pet_embeddings = pet_model.pet.embedding(batch_dict["neighbor_species"])

    torch.testing.assert_close(nativepet_embeddings, pet_embeddings)


def test_cartesian_transformer_compatibility():
    """Tests that the Cartesian transformer layers in the PET
    and NativePET models give the same predictions"""
    torch.manual_seed(0)
    nativepet_model, pet_model, systems = get_test_environment()

    nl_options = nativepet_model.requested_neighbor_lists()[0]
    batch_dict = systems_to_batch_dict(
        systems, nl_options, nativepet_model.atomic_types, None
    )
    batch_dict["input_messages"] = nativepet_model.embedding(
        batch_dict["neighbor_species"]
    )

    nativepet_cartesian_transformer = nativepet_model.gnn_layers[0]
    pet_cartesian_transformer = pet_model.pet.gnn_layers[0]

    edge_distances = torch.sqrt(torch.sum(batch_dict["x"] ** 2, dim=2) + 1e-16)
    cutoff_factors = cutoff_func(
        edge_distances, nativepet_model.cutoff, nativepet_model.cutoff_width
    )
    cutoff_factors[batch_dict["mask"]] = 0.0

    nativepet_output_node_embeddings, nativepet_output_messages = (
        nativepet_cartesian_transformer(
            input_messages=batch_dict["input_messages"],
            element_indices_nodes=batch_dict["central_species"],
            element_indices_neighbors=batch_dict["neighbor_species"],
            edge_vectors=batch_dict["x"],
            padding_mask=~batch_dict["mask"],
            edge_distances=edge_distances,
            cutoff_factors=cutoff_factors,
            use_manual_attention=False,
        )
    )
    pet_result = pet_cartesian_transformer(batch_dict)
    torch.testing.assert_close(
        nativepet_output_node_embeddings, pet_result["central_token"]
    )
    torch.testing.assert_close(nativepet_output_messages, pet_result["output_messages"])


def test_gnn_layers_compatibility():
    """Tests that the GNN layers in the PET and NativePET models
    give the same predictions"""
    nativepet_model, pet_model, systems = get_test_environment()

    nl_options = nativepet_model.requested_neighbor_lists()[0]
    nativepet_batch_dict = systems_to_batch_dict(
        systems, nl_options, nativepet_model.atomic_types, None
    )
    pet_batch_dict = systems_to_batch_dict(
        systems, nl_options, pet_model.atomic_types, None
    )

    neighbors_index = nativepet_batch_dict["neighbors_index"]
    neighbors_pos = nativepet_batch_dict["neighbors_pos"]

    nativepet_batch_dict["input_messages"] = nativepet_model.embedding(
        nativepet_batch_dict["neighbor_species"]
    )

    pet_batch_dict["input_messages"] = pet_model.pet.embedding(
        pet_batch_dict["neighbor_species"]
    )

    edge_distances = torch.sqrt(
        torch.sum(nativepet_batch_dict["x"] ** 2, dim=2) + 1e-15
    )
    cutoff_factors = cutoff_func(
        edge_distances, nativepet_model.cutoff, nativepet_model.cutoff_width
    )
    cutoff_factors[nativepet_batch_dict["mask"]] = 0.0

    for nativepet_gnn_layer, pet_gnn_layer in zip(
        nativepet_model.gnn_layers, pet_model.pet.gnn_layers
    ):
        output_node_embeddings, output_edge_embeddings = nativepet_gnn_layer(
            input_messages=nativepet_batch_dict["input_messages"],
            element_indices_nodes=nativepet_batch_dict["central_species"],
            element_indices_neighbors=nativepet_batch_dict["neighbor_species"],
            edge_vectors=nativepet_batch_dict["x"],
            padding_mask=~nativepet_batch_dict["mask"],
            edge_distances=edge_distances,
            cutoff_factors=cutoff_factors,
            use_manual_attention=False,
        )
        pet_result = pet_gnn_layer(pet_batch_dict)
        new_nativepet_input_messages = output_edge_embeddings[
            neighbors_index, neighbors_pos
        ]
        new_pet_input_messages = pet_result["output_messages"][
            neighbors_index, neighbors_pos
        ]

        nativepet_batch_dict["input_messages"] = 0.5 * (
            nativepet_batch_dict["input_messages"] + new_nativepet_input_messages
        )

        pet_batch_dict["input_messages"] = pet_model.pet.RESIDUAL_FACTOR * (
            pet_batch_dict["input_messages"] + new_pet_input_messages
        )

        torch.testing.assert_close(output_node_embeddings, pet_result["central_token"])
        torch.testing.assert_close(
            output_edge_embeddings, pet_result["output_messages"]
        )
        torch.testing.assert_close(
            nativepet_batch_dict["input_messages"], pet_batch_dict["input_messages"]
        )


def test_node_heads():
    """Tests that the heads in the PET and NativePET models
    give the same predictions"""
    nativepet_model, pet_model, systems = get_test_environment()

    nl_options = nativepet_model.requested_neighbor_lists()[0]
    batch_dict = systems_to_batch_dict(
        systems, nl_options, nativepet_model.atomic_types, None
    )
    batch_dict["input_messages"] = nativepet_model.embedding(
        batch_dict["neighbor_species"]
    )
    gnn_result = pet_model.pet.gnn_layers[0](batch_dict)
    pet_atomic_predictions = torch.zeros(1)
    nativepet_atomic_predictions = torch.zeros(1)

    pet_precitor_output = pet_model.pet.central_tokens_predictors[0](
        gnn_result["central_token"], batch_dict["central_species"]
    )
    pet_atomic_predictions = (
        pet_atomic_predictions + pet_precitor_output["atomic_predictions"]
    )
    nativepet_head_output = nativepet_model.node_heads["energy"][0](
        gnn_result["central_token"]
    )
    nativepet_atomic_predictions = (
        nativepet_atomic_predictions
        + nativepet_model.node_last_layers["energy"][0]["energy___0"](
            nativepet_head_output
        )
    )

    torch.testing.assert_close(nativepet_atomic_predictions, pet_atomic_predictions)


def test_edge_heads():
    """Tests that the bond heads in the PET and NativePET models
    give the same predictions"""
    nativepet_model, pet_model, systems = get_test_environment()

    nl_options = nativepet_model.requested_neighbor_lists()[0]
    batch_dict = systems_to_batch_dict(
        systems, nl_options, nativepet_model.atomic_types, None
    )
    x = batch_dict["x"]
    central_species = batch_dict["central_species"]
    mask = batch_dict["mask"]
    nums = batch_dict["nums"]

    lengths = torch.sqrt(torch.sum(x * x, dim=2) + 1e-16)
    multipliers = cutoff_func(lengths, pet_model.pet.R_CUT, pet_model.pet.CUTOFF_DELTA)
    multipliers[mask] = 0.0

    batch_dict["input_messages"] = pet_model.pet.embedding(
        batch_dict["neighbor_species"]
    )

    gnn_result = pet_model.pet.gnn_layers[0](batch_dict)

    pet_atomic_predictions = torch.zeros(1)
    nativepet_atomic_predictions = torch.zeros(1)

    pet_precitor_output = pet_model.pet.messages_bonds_predictors[0](
        gnn_result["output_messages"],
        mask,
        nums,
        central_species,
        multipliers,
    )

    pet_atomic_predictions = (
        pet_atomic_predictions + pet_precitor_output["atomic_predictions"]
    )

    nativepet_head_output = nativepet_model.edge_heads["energy"][0](
        gnn_result["output_messages"]
    )
    nativepet_last_layer_output = nativepet_model.edge_last_layers["energy"][0][
        "energy___0"
    ](nativepet_head_output)

    mask_expanded = mask[..., None].repeat(1, 1, nativepet_last_layer_output.shape[2])
    nativepet_last_layer_output = torch.where(
        mask_expanded, 0.0, nativepet_last_layer_output
    )
    nativepet_last_layer_output = nativepet_last_layer_output * multipliers[:, :, None]
    nativepet_last_layer_output = nativepet_last_layer_output.sum(dim=1)

    nativepet_atomic_predictions = (
        nativepet_atomic_predictions + nativepet_last_layer_output
    )

    torch.testing.assert_close(nativepet_atomic_predictions, pet_atomic_predictions)


def test_predictions_compatibility():
    """Tests that the predictions of the PET and NativePET models
    are the same"""
    nativepet_model, pet_model, systems = get_test_environment()

    outputs = {"energy": ModelOutput(per_atom=False)}

    nativepet_predictions = nativepet_model(systems, outputs)
    pet_predictions = pet_model(systems, outputs)

    torch.testing.assert_close(
        nativepet_predictions["energy"].block().values,
        pet_predictions["energy"].block().values,
    )


def test_positions_gradients_compatibility():
    """Tests that the gradients w.r.t positions of the
    PET and NativePET models are the same"""
    nativepet_model, pet_model, systems = get_test_environment()
    system = systems[0]

    outputs = {"energy": ModelOutput(per_atom=False)}

    nativepet_predictions = nativepet_model([system], outputs)

    nativepet_gradients = -torch.autograd.grad(
        nativepet_predictions["energy"].block().values[0][0],
        system.positions,
        torch.ones_like(nativepet_predictions["energy"].block().values[0][0]),
        create_graph=True,
        retain_graph=True,
    )[0]

    nl_options = nativepet_model.requested_neighbor_lists()[0]
    batch_dict = systems_to_batch_dict(
        [system], nl_options, nativepet_model.atomic_types, None
    )
    x = batch_dict["x"]
    x.requires_grad_(True)

    pet_predictions = pet_model.pet(batch_dict)["prediction"]

    pet_grads_wrt_x = torch.autograd.grad(
        pet_predictions,
        x,
        grad_outputs=torch.ones_like(pet_predictions),
        create_graph=True,
        retain_graph=True,
    )[0]

    neighbors_index = batch_dict["neighbors_index"]  # .transpose(0, 1)
    neighbors_pos = batch_dict["neighbors_pos"]
    grads_messaged = pet_grads_wrt_x[neighbors_index, neighbors_pos]
    pet_grads_wrt_x[batch_dict["mask"]] = 0.0

    grads_messaged[batch_dict["mask"]] = 0.0
    first = pet_grads_wrt_x.sum(dim=1)
    second = grads_messaged.sum(dim=1)
    pet_gradients = first - second

    torch.testing.assert_close(
        nativepet_gradients,
        pet_gradients,
    )


def test_energy_loss_grads_compatibility():
    """Tests that the gradients of the energy loss produced by
    PET and NativePET models are the same"""
    nativepet_model, pet_model, systems = get_test_environment()

    outputs = {"energy": ModelOutput(per_atom=False)}

    nativepet_predictions = nativepet_model(systems, outputs)
    nativepet_loss = (
        nativepet_predictions["energy"].block().values.sum()
    )  # Artificial loss
    nativepet_loss.backward()

    pet_predictions = pet_model(systems, outputs)
    pet_loss = pet_predictions["energy"].block().values.sum()  # Artificial loss
    pet_loss.backward()

    nativepet_params = dict(nativepet_model.named_parameters())
    pet_params = dict(pet_model.named_parameters())

    for key, nativepet_param in nativepet_params.items():
        if "last_layers" in key:
            continue
        pet_key = "pet." + key
        if "heads" in key:
            pet_key = pet_key.replace(".energy", "")
        if pet_key in pet_params:
            pet_param = pet_params.pop(pet_key)
            torch.testing.assert_close(nativepet_param.grad, pet_param.grad)


def test_forces_loss_grads_compatibility():
    """Tests that the gradients of the loss produced by PET and NativePET
    models are the same"""
    nativepet_model, pet_model, systems = get_test_environment()
    system = systems[0]

    outputs = {"energy": ModelOutput(per_atom=False)}

    nativepet_predictions = nativepet_model([system], outputs)

    nativepet_gradients = -torch.autograd.grad(
        nativepet_predictions["energy"].block().values[0][0],
        system.positions,
        torch.ones_like(nativepet_predictions["energy"].block().values[0][0]),
        create_graph=True,
        retain_graph=True,
    )[0]

    nl_options = nativepet_model.requested_neighbor_lists()[0]
    batch_dict = systems_to_batch_dict(
        [system], nl_options, nativepet_model.atomic_types, None
    )
    x = batch_dict["x"]
    x.requires_grad_(True)

    pet_predictions = pet_model.pet(batch_dict)["prediction"]

    pet_grads_wrt_x = torch.autograd.grad(
        pet_predictions,
        x,
        grad_outputs=torch.ones_like(pet_predictions),
        create_graph=True,
        retain_graph=True,
    )[0]

    neighbors_index = batch_dict["neighbors_index"]  # .transpose(0, 1)
    neighbors_pos = batch_dict["neighbors_pos"]
    grads_messaged = pet_grads_wrt_x[neighbors_index, neighbors_pos]
    pet_grads_wrt_x[batch_dict["mask"]] = 0.0

    grads_messaged[batch_dict["mask"]] = 0.0
    first = pet_grads_wrt_x.sum(dim=1)
    second = grads_messaged.sum(dim=1)
    pet_gradients = first - second

    nativepet_loss = nativepet_gradients.sum()
    nativepet_loss.backward()

    pet_loss = pet_gradients.sum()
    pet_loss.backward()

    nativepet_params = dict(nativepet_model.named_parameters())
    pet_params = dict(pet_model.named_parameters())

    for key, nativepet_param in nativepet_params.items():
        if "last_layers" in key:
            continue
        pet_key = "pet." + key
        if "heads" in key:
            pet_key = pet_key.replace(".energy", "")
        if pet_key in pet_params:
            pet_param = pet_params.pop(pet_key)
            torch.testing.assert_close(nativepet_param.grad, pet_param.grad)


def test_last_layer_features_compatibility():
    nativepet_model, pet_model, systems = get_test_environment()
    system = systems[0]
    outputs = {
        "energy": ModelOutput(per_atom=False),
        "mtt::aux::energy_last_layer_features": ModelOutput(per_atom=False),
    }

    nativepet_predictions = nativepet_model([system], outputs)
    pet_predictions = pet_model([system], outputs)

    nativepet_last_layer_features = nativepet_predictions[
        "mtt::aux::energy_last_layer_features"
    ]

    pet_last_layer_features = pet_predictions["mtt::aux::energy_last_layer_features"]

    assert metatensor.torch.allclose(
        nativepet_last_layer_features, pet_last_layer_features, atol=1e-6
    )


def test_pet_mad_model_compatibility(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)

    path = "https://huggingface.co/lab-cosmo/pet-mad/resolve/main/models/pet-mad-latest.ckpt"

    if urlparse(path).scheme:
        path, _ = urlretrieve(path)

    pet_model = PET.load_checkpoint(path).eval()

    nativepet_checkpoint = torch.load(path, map_location="cpu", weights_only=False)
    nativepet_checkpoint = convert_checkpoint_from_legacy_pet(nativepet_checkpoint)
    torch.save(
        nativepet_checkpoint,
        "nativepet_checkpoint.ckpt",
    )
    nativepet_model = NativePET.load_checkpoint("nativepet_checkpoint.ckpt").eval()

    systems_1 = read_systems(DATASET_PATH)[:5]
    systems_2 = read_systems(DATASET_WITH_FORCES_PATH)[:5]
    systems = systems_1 + systems_2
    for system in systems:
        system.positions.requires_grad_(True)
        get_system_with_neighbor_lists(
            system, nativepet_model.requested_neighbor_lists()
        )
    systems = [system.to(torch.float32) for system in systems]

    outputs = {"energy": ModelOutput(per_atom=False)}

    nativepet_predictions = nativepet_model(systems, outputs)
    pet_predictions = pet_model(systems, outputs)

    nativepet_gradients = compute_gradient(
        nativepet_predictions["energy"].block().values,
        [system.positions for system in systems],
        is_training=True,
    )

    pet_gradients = compute_gradient(
        pet_predictions["energy"].block().values,
        [system.positions for system in systems],
        is_training=True,
    )

    torch.testing.assert_close(
        nativepet_predictions["energy"].block().values,
        pet_predictions["energy"].block().values,
    )

    torch.testing.assert_close(nativepet_gradients, pet_gradients)



import pytest
import torch
from metatensor.torch.atomistic import (
    MetatensorAtomisticModel,
    ModelCapabilities,
    ModelEvaluationOptions,
    ModelMetadata,
    ModelOutput,
    NeighborListOptions,
    systems_to_torch,
)

from metatrain.pet import PET as WrappedPET
from metatrain.pet.modules.data_preparation import get_pyg_graphs
from metatrain.pet.modules.hypers import Hypers
from metatrain.pet.modules.pet import PET
from metatrain.pet.utils import systems_to_batch_dict
from metatrain.utils.architectures import get_default_hypers
from metatrain.utils.data import DatasetInfo
from metatrain.utils.data.readers.ase import read
from metatrain.utils.data.target_info import get_energy_target_info
from metatrain.utils.neighbor_lists import get_system_with_neighbor_lists

from . import DATASET_PATH


DEFAULT_HYPERS = get_default_hypers("pet")


def check_batch_dict_consistency(ref_batch, trial_batch):
    ref_mask = ref_batch["mask"]
    trial_mask = trial_batch["mask"]
    assert torch.all(ref_mask == trial_mask)
    mask = ref_mask is False

    for key in ref_batch:
        if key == "x":
            torch.testing.assert_close(
                ref_batch["x"].flatten().sort()[0],
                trial_batch["x"].flatten().sort()[0],
                rtol=1e-6,
                atol=1e-5,
            )
        elif key in ("central_species", "mask", "nums", "batch"):
            assert torch.all(ref_batch[key] == trial_batch[key])
        else:
            ref_unique, ref_counts = ref_batch[key][mask].unique(return_counts=True)
            trial_unique, trial_counts = trial_batch[key][mask].unique(
                return_counts=True
            )
            assert torch.all(ref_unique == trial_unique)
            assert torch.all(ref_counts == trial_counts)


@pytest.mark.parametrize("cutoff", [0.25, 5.0])
def test_batch_dicts_compatibility(cutoff):
    """Tests that the batch dict computed with internal MTM routines
    is consitent with PET implementation."""

    structure = read(DATASET_PATH)
    atomic_types = sorted(set(structure.numbers))
    system = systems_to_torch(structure)
    options = NeighborListOptions(cutoff=cutoff, full_list=True, strict=True)
    system = get_system_with_neighbor_lists(system, [options])

    ARCHITECTURAL_HYPERS = Hypers(DEFAULT_HYPERS["model"])
    batch = get_pyg_graphs(
        [structure],
        atomic_types,
        cutoff,
        ARCHITECTURAL_HYPERS.USE_ADDITIONAL_SCALAR_ATTRIBUTES,
        ARCHITECTURAL_HYPERS.USE_LONG_RANGE,
        ARCHITECTURAL_HYPERS.K_CUT,
        ARCHITECTURAL_HYPERS.N_TARGETS > 1,
        ARCHITECTURAL_HYPERS.TARGET_INDEX_KEY,
    )[0]
    ref_batch_dict = {
        "x": batch.x,
        "central_species": batch.central_species,
        "neighbor_species": batch.neighbor_species,
        "mask": batch.mask,
        "batch": torch.tensor([0] * len(batch.central_species)),
        "nums": batch.nums,
        "neighbors_index": batch.neighbors_index.transpose(0, 1),
        "neighbors_pos": batch.neighbors_pos,
    }
    trial_batch_dict = systems_to_batch_dict([system], options, atomic_types, None)
    check_batch_dict_consistency(ref_batch_dict, trial_batch_dict)


@pytest.mark.parametrize("cutoff", [0.25, 5.0])
def test_predictions_compatibility(cutoff):
    """Tests that predictions of the MTM implemetation of PET
    are consistent with the predictions of the original PET implementation."""

    structure = read(DATASET_PATH)

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=structure.numbers,
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )
    capabilities = ModelCapabilities(
        length_unit="Angstrom",
        atomic_types=dataset_info.atomic_types,
        outputs={
            "energy": ModelOutput(
                quantity="energy",
                unit="eV",
            )
        },
        interaction_range=DEFAULT_HYPERS["model"]["N_GNN_LAYERS"]
        * DEFAULT_HYPERS["model"]["R_CUT"],
        dtype="float32",
        supported_devices=["cpu", "cuda"],
    )
    hypers = DEFAULT_HYPERS["model"]
    hypers["R_CUT"] = cutoff
    model = WrappedPET(DEFAULT_HYPERS["model"], dataset_info)
    ARCHITECTURAL_HYPERS = Hypers(model.hypers)
    raw_pet = PET(ARCHITECTURAL_HYPERS, 0.0, len(dataset_info.atomic_types))
    model.set_trained_model(raw_pet)

    system = systems_to_torch(structure)
    options = NeighborListOptions(cutoff=cutoff, full_list=True, strict=True)
    system = get_system_with_neighbor_lists(system, [options])

    evaluation_options = ModelEvaluationOptions(
        length_unit=capabilities.length_unit,
        outputs=capabilities.outputs,
    )

    model = MetatensorAtomisticModel(model.eval(), ModelMetadata(), capabilities)
    mtm_pet_prediction = (
        model(
            [system],
            evaluation_options,
            check_consistency=False,
        )["energy"]
        .block()
        .values
    )

    ARCHITECTURAL_HYPERS = Hypers(DEFAULT_HYPERS["model"])
    batch = get_pyg_graphs(
        [structure],
        dataset_info.atomic_types,
        cutoff,
        ARCHITECTURAL_HYPERS.USE_ADDITIONAL_SCALAR_ATTRIBUTES,
        ARCHITECTURAL_HYPERS.USE_LONG_RANGE,
        ARCHITECTURAL_HYPERS.K_CUT,
        ARCHITECTURAL_HYPERS.N_TARGETS > 1,
        ARCHITECTURAL_HYPERS.TARGET_INDEX_KEY,
    )[0]

    batch_dict = {
        "x": batch.x,
        "central_species": batch.central_species,
        "neighbor_species": batch.neighbor_species,
        "mask": batch.mask,
        "batch": torch.tensor([0] * len(batch.central_species)),
        "nums": batch.nums,
        "neighbors_index": batch.neighbors_index.transpose(0, 1),
        "neighbors_pos": batch.neighbors_pos,
    }

    pet = model.module.pet

    pet_prediction = pet.forward(batch_dict)["prediction"]

    torch.testing.assert_close(
        mtm_pet_prediction, pet_prediction.sum(dim=0, keepdim=True)
    )



from typing import List, Tuple

import metatensor.torch
import torch
from metatensor.torch import Labels, TensorBlock, TensorMap
from metatensor.torch.atomistic import System
from omegaconf import DictConfig

from ..target_info import TargetInfo, get_energy_target_info, get_generic_target_info


def read_systems(filename: str) -> List[System]:
    """Read system information using metatensor.

    :param filename: name of the file to read

    :raises NotImplementedError: Serialization of systems is not yet
        available in metatensor.
    """
    raise NotImplementedError("Reading metatensor systems is not yet implemented.")


def _wrapped_metatensor_read(filename) -> TensorMap:
    try:
        return metatensor.torch.load(filename)
    except Exception as e:
        raise ValueError(f"Failed to read '{filename}' with torch: {e}") from e


def read_energy(target: DictConfig) -> Tuple[TensorMap, TargetInfo]:
    tensor_map = _wrapped_metatensor_read(target["read_from"])

    if len(tensor_map) != 1:
        raise ValueError("Energy TensorMaps should have exactly one block.")

    add_position_gradients = target["forces"]
    add_strain_gradients = target["stress"] or target["virial"]
    target_info = get_energy_target_info(
        target, add_position_gradients, add_strain_gradients
    )

    # now check all the expected metadata (from target_info.layout) matches
    # the actual metadata in the tensor maps
    _check_tensor_map_metadata(tensor_map, target_info.layout)

    selections = [
        Labels(
            names=["system"],
            values=torch.tensor([[int(i)]]),
        )
        for i in torch.unique(
            torch.concatenate(
                [block.samples.column("system") for block in tensor_map.blocks()]
            )
        )
    ]
    tensor_maps = metatensor.torch.split(tensor_map, "samples", selections)
    return tensor_maps, target_info


def read_generic(target: DictConfig) -> Tuple[List[TensorMap], TargetInfo]:
    tensor_map = _wrapped_metatensor_read(target["read_from"])

    for block in tensor_map.blocks():
        if len(block.gradients_list()) > 0:
            raise ValueError("Only energy targets can have gradient blocks.")

    target_info = get_generic_target_info(target)
    _check_tensor_map_metadata(tensor_map, target_info.layout)

    # make sure that the properties of the target_info.layout also match the
    # actual properties of the tensor maps
    target_info.layout = _empty_tensor_map_like(tensor_map)

    selections = [
        Labels(
            names=["system"],
            values=torch.tensor([[int(i)]]),
        )
        for i in torch.unique(tensor_map.block(0).samples.column("system"))
    ]
    tensor_maps = metatensor.torch.split(tensor_map, "samples", selections)
    return tensor_maps, target_info


def _check_tensor_map_metadata(tensor_map: TensorMap, layout: TensorMap):
    if tensor_map.keys != layout.keys:
        raise ValueError(
            f"Unexpected keys in metatensor targets: "
            f"expected: {layout.keys} "
            f"actual: {tensor_map.keys}"
        )
    for key in layout.keys:
        block = tensor_map.block(key)
        block_from_layout = layout.block(key)
        if block.samples.names != block_from_layout.samples.names:
            raise ValueError(
                f"Unexpected samples in metatensor targets: "
                f"expected: {block_from_layout.samples.names} "
                f"actual: {block.samples.names}"
            )
        if block.components != block_from_layout.components:
            raise ValueError(
                f"Unexpected components in metatensor targets: "
                f"expected: {block_from_layout.components} "
                f"actual: {block.components}"
            )
        # the properties can be different from those of the default `TensorMap`
        # given by `get_generic_target_info`, so we don't check them
        if set(block.gradients_list()) != set(block_from_layout.gradients_list()):
            raise ValueError(
                f"Unexpected gradients in metatensor targets: "
                f"expected: {block_from_layout.gradients_list()} "
                f"actual: {block.gradients_list()}"
            )
        for name in block_from_layout.gradients_list():
            gradient_block = block.gradient(name)
            gradient_block_from_layout = block_from_layout.gradient(name)
            if gradient_block.labels.names != gradient_block_from_layout.labels.names:
                raise ValueError(
                    f"Unexpected samples in metatensor targets "
                    f"for `{name}` gradient block: "
                    f"expected: {gradient_block_from_layout.labels.names} "
                    f"actual: {gradient_block.labels.names}"
                )
            if gradient_block.components != gradient_block_from_layout.components:
                raise ValueError(
                    f"Unexpected components in metatensor targets "
                    f"for `{name}` gradient block: "
                    f"expected: {gradient_block_from_layout.components} "
                    f"actual: {gradient_block.components}"
                )


def _empty_tensor_map_like(tensor_map: TensorMap) -> TensorMap:
    new_keys = tensor_map.keys
    new_blocks: List[TensorBlock] = []
    for block in tensor_map.blocks():
        new_block = _empty_tensor_block_like(block)
        new_blocks.append(new_block)
    return TensorMap(keys=new_keys, blocks=new_blocks)


def _empty_tensor_block_like(tensor_block: TensorBlock) -> TensorBlock:
    new_block = TensorBlock(
        values=torch.empty(
            (0,) + tensor_block.values.shape[1:],
            dtype=torch.float64,  # metatensor can't serialize otherwise
            device=tensor_block.values.device,
        ),
        samples=Labels(
            names=tensor_block.samples.names,
            values=torch.empty(
                (0, tensor_block.samples.values.shape[1]),
                dtype=tensor_block.samples.values.dtype,
                device=tensor_block.samples.values.device,
            ),
        ),
        components=tensor_block.components,
        properties=tensor_block.properties,
    )
    for gradient_name, gradient in tensor_block.gradients():
        new_block.add_gradient(gradient_name, _empty_tensor_block_like(gradient))
    return new_block



from .dataset_to_ase import dataset_to_ase
from .load_raw_pet_model import load_raw_pet_model
from .systems_to_batch_dict import systems_to_batch_dict
from .update_hypers import update_hypers
from .update_state_dict import update_state_dict


__all__ = [
    "systems_to_batch_dict",
    "dataset_to_ase",
    "update_hypers",
    "update_state_dict",
    "load_raw_pet_model",
]



import torch


class DistributedDataParallel(torch.nn.parallel.DistributedDataParallel):
    """
    DistributedDataParallel wrapper that inherits from
    :py:class`torch.nn.parallel.DistributedDataParallel`
    and adds the capabilities attribute to it.

    :param module: The module to be parallelized.
    :param args: Arguments to be passed to the parent class.
    :param kwargs: Keyword arguments to be passed to the parent class
    """

    def __init__(self, module: torch.nn.Module, *args, **kwargs):
        super(DistributedDataParallel, self).__init__(module, *args, **kwargs)
        self.outputs = module.outputs



from .model import SoapBpnn
from .trainer import Trainer


__model__ = SoapBpnn
__trainer__ = Trainer

__authors__ = [
    ("Filippo Bigi <filippo.bigi@epfl.ch>", "@frostedoyster"),
]

__maintainers__ = [
    ("Filippo Bigi <filippo.bigi@epfl.ch>", "@frostedoyster"),
]



from typing import Dict, List

import numpy as np
import torch

from ..modules.hypers import Hypers
from ..modules.pet import PET, SelfContributionsWrapper
from .fine_tuning import LoRAWrapper
from .update_state_dict import update_state_dict


def load_raw_pet_model(
    state_dict: Dict,
    hypers: Dict,
    atomic_types: List,
    self_contributions: np.ndarray,
    **kwargs,
) -> "SelfContributionsWrapper":
    """Creates a raw PET model instance."""

    ARCHITECTURAL_HYPERS = Hypers(hypers)

    ARCHITECTURAL_HYPERS.D_OUTPUT = 1  # type: ignore
    ARCHITECTURAL_HYPERS.TARGET_AGGREGATION = "sum"  # type: ignore
    ARCHITECTURAL_HYPERS.TARGET_TYPE = "atomic"  # type: ignore

    raw_pet = PET(ARCHITECTURAL_HYPERS, 0.0, len(atomic_types))
    if "use_lora_peft" in kwargs and kwargs["use_lora_peft"] is True:
        lora_rank = kwargs["lora_rank"]
        lora_alpha = kwargs["lora_alpha"]
        raw_pet = LoRAWrapper(raw_pet, lora_rank, lora_alpha)

    new_state_dict = update_state_dict(state_dict)
    dtype = next(iter(new_state_dict.values())).dtype
    raw_pet.to(dtype).load_state_dict(new_state_dict)
    if isinstance(self_contributions, torch.Tensor):
        self_contributions = self_contributions.cpu().numpy()
    wrapper = SelfContributionsWrapper(raw_pet, self_contributions)

    return wrapper



from typing import List

import ase.neighborlist
import numpy as np
import torch
import vesin
from metatensor.torch import Labels, TensorBlock
from metatensor.torch.atomistic import (
    NeighborListOptions,
    System,
)

from .data.system_to_ase import system_to_ase


def get_requested_neighbor_lists(
    module: torch.nn.Module,
) -> List[NeighborListOptions]:
    """Get the neighbor lists requested by a module and its children.

    :param module: The module for which to get the requested neighbor lists.

    :return: A list of `NeighborListOptions` objects requested by the module.
    """
    requested: List[NeighborListOptions] = []
    _get_requested_neighbor_lists_in_place(
        module=module,
        module_name="",
        requested=requested,
    )
    return requested


def _get_requested_neighbor_lists_in_place(
    module: torch.nn.Module,
    module_name: str,
    requested: List[NeighborListOptions],
):
    # copied from
    # metatensor/python/metatensor-torch/metatensor/torch/atomistic/model.py
    # and just removed the length units

    if hasattr(module, "requested_neighbor_lists"):
        for new_options in module.requested_neighbor_lists():
            new_options.add_requestor(module_name)

            already_requested = False
            for existing in requested:
                if existing == new_options:
                    already_requested = True
                    for requestor in new_options.requestors():
                        existing.add_requestor(requestor)

            if not already_requested:
                requested.append(new_options)

    for child_name, child in module.named_children():
        _get_requested_neighbor_lists_in_place(
            module=child,
            module_name=module_name + "." + child_name,
            requested=requested,
        )


def get_system_with_neighbor_lists(
    system: System, neighbor_lists: List[NeighborListOptions]
) -> System:
    """Attaches neighbor lists to a `System` object.

    :param system: The system for which to calculate neighbor lists.
    :param neighbor_lists: A list of `NeighborListOptions` objects,
        each of which specifies the parameters for a neighbor list.

    :return: The `System` object with the neighbor lists added.
    """
    # Convert the system to an ASE atoms object
    atoms = system_to_ase(system)

    # Compute the neighbor lists
    for options in neighbor_lists:
        if options not in system.known_neighbor_lists():
            neighbor_list = _compute_single_neighbor_list(atoms, options).to(
                device=system.device, dtype=system.dtype
            )
            system.add_neighbor_list(options, neighbor_list)

    return system


def _compute_single_neighbor_list(
    atoms: ase.Atoms, options: NeighborListOptions
) -> TensorBlock:
    # Computes a single neighbor list for an ASE atoms object
    # (as in metatensor.torch.atomistic)

    if np.all(atoms.pbc) or np.all(~atoms.pbc):
        nl_i, nl_j, nl_S, nl_D = vesin.ase_neighbor_list(
            "ijSD",
            atoms,
            cutoff=options.cutoff,
        )
    else:
        # this is not implemented in vesin, so we use ASE
        nl_i, nl_j, nl_S, nl_D = ase.neighborlist.neighbor_list(
            "ijSD",
            atoms,
            cutoff=options.cutoff,
        )

    # The pair selection code here below avoids a relatively slow loop over
    # all pairs to improve performance
    reject_condition = (
        # we want a half neighbor list, so drop all duplicated neighbors
        (nl_j < nl_i)
        | (
            (nl_i == nl_j)
            & (
                # only create pairs with the same atom twice if the pair spans more
                # than one unit cell
                ((nl_S[:, 0] == 0) & (nl_S[:, 1] == 0) & (nl_S[:, 2] == 0))
                |
                # When creating pairs between an atom and one of its periodic images,
                # the code generates multiple redundant pairs
                # (e.g. with shifts 0 1 1 and 0 -1 -1); and we want to only keep one of
                # these. We keep the pair in the positive half plane of shifts.
                (
                    (nl_S.sum(axis=1) < 0)
                    | (
                        (nl_S.sum(axis=1) == 0)
                        & ((nl_S[:, 2] < 0) | ((nl_S[:, 2] == 0) & (nl_S[:, 1] < 0)))
                    )
                )
            )
        )
    )
    selected = np.logical_not(reject_condition)
    n_pairs = np.sum(selected)

    if options.full_list:
        distances = np.empty((2 * n_pairs, 3), dtype=np.float64)
        samples = np.empty((2 * n_pairs, 5), dtype=np.int32)
    else:
        distances = np.empty((n_pairs, 3), dtype=np.float64)
        samples = np.empty((n_pairs, 5), dtype=np.int32)

    samples[:n_pairs, 0] = nl_i[selected]
    samples[:n_pairs, 1] = nl_j[selected]
    samples[:n_pairs, 2:] = nl_S[selected]

    distances[:n_pairs] = nl_D[selected]

    if options.full_list:
        samples[n_pairs:, 0] = nl_j[selected]
        samples[n_pairs:, 1] = nl_i[selected]
        samples[n_pairs:, 2:] = -nl_S[selected]

        distances[n_pairs:] = -nl_D[selected]

    distances = torch.from_numpy(distances)
    return TensorBlock(
        values=distances.reshape(-1, 3, 1),
        samples=Labels(
            names=[
                "first_atom",
                "second_atom",
                "cell_shift_a",
                "cell_shift_b",
                "cell_shift_c",
            ],
            values=torch.from_numpy(samples),
        ),
        components=[Labels.range("xyz", 3)],
        properties=Labels.range("distance", 1),
    )



"""Modules to allow SOAP-BPNN to fit arbitrary spherical tensor targets."""

import copy
from typing import Dict, List, Optional

import featomic.torch
import metatensor.torch
import numpy as np
import torch
import wigners
from metatensor.torch import Labels
from metatensor.torch.atomistic import System
from metatensor.torch.learn.nn import Linear as LinearMap


class VectorBasis(torch.nn.Module):
    """
    This module creates a basis of 3 vectors for each atomic environment.

    In practice, this is done by contracting a l=1 spherical expansion.
    """

    def __init__(self, atomic_types, soap_hypers) -> None:
        super().__init__()
        self.atomic_types = atomic_types
        soap_vector_hypers = copy.deepcopy(soap_hypers)
        soap_vector_hypers["basis"]["max_angular"] = 1
        self.soap_calculator = featomic.torch.SphericalExpansion(**soap_vector_hypers)
        self.neighbor_species_labels = Labels(
            names=["neighbor_type"],
            values=torch.tensor(self.atomic_types).reshape(-1, 1),
        )
        self.contraction = LinearMap(
            in_keys=Labels(
                names=["o3_lambda", "o3_sigma", "center_type"],
                values=torch.stack(
                    [
                        torch.tensor([1] * len(self.atomic_types)),
                        torch.tensor([1] * len(self.atomic_types)),
                        torch.tensor(self.atomic_types),
                    ],
                    dim=1,
                ),
            ),
            in_features=(soap_vector_hypers["basis"]["radial"]["max_radial"] + 1)
            * len(self.atomic_types),
            out_features=3,
            bias=False,
            out_properties=[Labels.range("basis", 3) for _ in self.atomic_types],
        )
        # this optimizable basis seems to work much better than a fixed one

    def forward(
        self, systems: List[System], selected_atoms: Optional[Labels]
    ) -> torch.Tensor:
        device = systems[0].positions.device
        if self.neighbor_species_labels.device != device:
            self.neighbor_species_labels = self.neighbor_species_labels.to(device)

        spherical_expansion = self.soap_calculator(
            systems, selected_samples=selected_atoms
        )

        # by calling keys_to_samples and keys_to_properties in the same order as they
        # are called in the main model, we should ensure that the order of the samples
        # is the same
        spherical_expansion = spherical_expansion.keys_to_properties(
            self.neighbor_species_labels
        )

        # drop all L=0 blocks
        spherical_expansion = metatensor.torch.drop_blocks(
            spherical_expansion,
            keys=Labels(
                ["o3_lambda", "o3_sigma"], torch.tensor([[0, 1]], device=device)
            ),
        )

        basis_vectors = self.contraction(spherical_expansion)
        basis_vectors = basis_vectors.keys_to_samples("center_type")

        basis_vectors_as_tensor = basis_vectors.block({"o3_lambda": 1}).values
        return basis_vectors_as_tensor  # [n_atoms, 3(yzx), 3]


class TensorBasis(torch.nn.Module):
    """
    Creates a basis of spherical tensors for each atomic environment. Internally, it
    uses one (for proper tensors) or two (for pseudotensors) VectorBasis objects to
    build a basis of 3 vectors.
    """

    cgs: Dict[str, torch.Tensor]  # torchscript needs this

    def __init__(self, atomic_types, soap_hypers, o3_lambda, o3_sigma) -> None:
        super().__init__()

        self.o3_lambda = o3_lambda
        self.o3_sigma = o3_sigma
        if self.o3_lambda > 0:
            self.vector_basis = VectorBasis(atomic_types, soap_hypers)
        else:
            self.vector_basis = FakeVectorBasis()  # needed to make torchscript work
        if self.o3_sigma == -1:
            self.vector_basis_pseudotensor = VectorBasis(atomic_types, soap_hypers)
        else:
            self.vector_basis_pseudotensor = FakeVectorBasis()  # make torchscript work

        if self.o3_lambda > 1:
            try:
                import sphericart.torch
            except ImportError:
                raise ImportError(
                    "To use spherical tensors with lambda > 1 with SOAP-BPNN, please "
                    "install the `sphericart-torch` package."
                )
            self.spherical_hamonics_calculator = sphericart.torch.SphericalHarmonics(
                l_max=self.o3_lambda
            )
        else:
            # needed to make torchscript work
            self.spherical_hamonics_calculator = torch.nn.Identity()

        if self.o3_lambda > 1 or self.o3_sigma == -1:
            self.cgs = {
                f"{l1}_{l2}_{L}": cg_tensor
                for (l1, l2, L), cg_tensor in get_cg_coefficients(
                    max(self.o3_lambda, 1)  # need at least 1 for pseudoscalar case
                )._cgs.items()
            }
        else:
            # needed to make torchscript work
            self.cgs = {}  # type: ignore

    def forward(
        self, systems: List[System], selected_atoms: Optional[Labels]
    ) -> torch.Tensor:
        # transfer cg dict to device and dtype if needed
        device = systems[0].positions.device
        dtype = systems[0].positions.dtype
        for k, v in self.cgs.items():
            if v.device != device or v.dtype != dtype:
                self.cgs[k] = v.to(device, dtype)

        if selected_atoms is None:
            num_atoms = sum(len(system) for system in systems)
        else:
            num_atoms = len(selected_atoms)

        if self.o3_lambda == 0:
            basis = torch.ones(
                (num_atoms, 1, 1),
                device=device,
                dtype=dtype,
            )
        elif self.o3_lambda == 1:
            basis = self.vector_basis(systems, selected_atoms)
            basis = basis / torch.sqrt(
                torch.sum(torch.square(basis), dim=1, keepdim=True)
            )
        elif self.o3_lambda == 2:
            basis = torch.empty(
                (num_atoms, 5, 5),
                device=device,
                dtype=dtype,
            )
            vector_basis = self.vector_basis(systems, selected_atoms)
            # vector_basis is [n_atoms, 3(yzx), 3]
            vector_1_xyz = vector_basis[:, [2, 0, 1], 0]
            vector_2_xyz = vector_basis[:, [2, 0, 1], 1]
            basis[:, :, 0] = self.spherical_hamonics_calculator(vector_1_xyz)[:, 4:]
            basis[:, :, 1] = self.spherical_hamonics_calculator(vector_2_xyz)[:, 4:]
            vector_1_spherical = vector_basis[:, :, 0]
            vector_2_spherical = vector_basis[:, :, 1]
            vector_3_spherical = vector_basis[:, :, 2]
            vector_1_spherical = vector_1_spherical / torch.sqrt(
                torch.sum(torch.square(vector_1_spherical), dim=-1, keepdim=True)
            )
            vector_2_spherical = vector_2_spherical / torch.sqrt(
                torch.sum(torch.square(vector_2_spherical), dim=-1, keepdim=True)
            )
            vector_3_spherical = vector_3_spherical / torch.sqrt(
                torch.sum(torch.square(vector_3_spherical), dim=-1, keepdim=True)
            )
            basis[:, :, 2] = cg_combine(
                vector_1_spherical, vector_2_spherical, self.cgs["1_1_2"]
            )
            basis[:, :, 3] = cg_combine(
                vector_1_spherical, vector_3_spherical, self.cgs["1_1_2"]
            )
            basis[:, :, 4] = cg_combine(
                vector_2_spherical, vector_3_spherical, self.cgs["1_1_2"]
            )
        else:  # self.o3_lambda > 2
            basis = torch.empty(
                (
                    num_atoms,
                    2 * self.o3_lambda + 1,
                    2 * self.o3_lambda + 1,
                ),
                device=device,
                dtype=dtype,
            )
            vector_basis = self.vector_basis(systems, selected_atoms)
            # vector_basis is [n_atoms, 3(yzx), 3]
            vector_1_xyz = vector_basis[:, [2, 0, 1], 0]
            vector_2_xyz = vector_basis[:, [2, 0, 1], 1]
            vector_3_spherical = vector_basis[:, :, 2]
            vector_3_spherical = vector_3_spherical / torch.sqrt(
                torch.sum(torch.square(vector_3_spherical), dim=-1, keepdim=True)
            )
            sh_1 = self.spherical_hamonics_calculator(vector_1_xyz)
            sh_2 = self.spherical_hamonics_calculator(vector_2_xyz)
            for lam in range(self.o3_lambda + 1):
                basis[:, :, lam] = cg_combine(
                    sh_1[:, lam * lam : (lam + 1) * (lam + 1)],
                    sh_2[
                        :,
                        (self.o3_lambda - lam) * (self.o3_lambda - lam) : (
                            (self.o3_lambda - lam) + 1
                        )
                        * ((self.o3_lambda - lam) + 1),
                    ],
                    self.cgs[
                        str(lam)
                        + "_"
                        + str(self.o3_lambda - lam)
                        + "_"
                        + str(self.o3_lambda)
                    ],
                )
            for lam in range(self.o3_lambda):
                basis[:, :, self.o3_lambda + 1 + lam] = cg_combine(
                    cg_combine(
                        sh_1[:, lam * lam : (lam + 1) * (lam + 1)],
                        sh_2[
                            :,
                            (self.o3_lambda - lam - 1) * (self.o3_lambda - lam - 1) : (
                                (self.o3_lambda - lam - 1) + 1
                            )
                            * ((self.o3_lambda - lam - 1) + 1),
                        ],
                        self.cgs[
                            str(lam)
                            + "_"
                            + str(self.o3_lambda - lam - 1)
                            + "_"
                            + str(self.o3_lambda - 1)
                        ],
                    ),
                    vector_3_spherical,
                    self.cgs[str(self.o3_lambda - 1) + "_1_" + str(self.o3_lambda)],
                )

        if self.o3_sigma == -1:
            # multiply by pseudoscalar
            vector_basis_pseudotensor = self.vector_basis_pseudotensor(
                systems, selected_atoms
            )
            vector_1_spherical = vector_basis_pseudotensor[:, :, 0]
            vector_2_spherical = vector_basis_pseudotensor[:, :, 1]
            vector_3_spherical = vector_basis_pseudotensor[:, :, 2]
            vector_1_spherical = vector_1_spherical / torch.sqrt(
                torch.sum(torch.square(vector_1_spherical), dim=-1, keepdim=True)
            )
            vector_2_spherical = vector_2_spherical / torch.sqrt(
                torch.sum(torch.square(vector_2_spherical), dim=-1, keepdim=True)
            )
            vector_3_spherical = vector_3_spherical / torch.sqrt(
                torch.sum(torch.square(vector_3_spherical), dim=-1, keepdim=True)
            )
            pseudoscalar = cg_combine(
                cg_combine(vector_1_spherical, vector_2_spherical, self.cgs["1_1_1"]),
                vector_3_spherical,
                self.cgs["1_1_0"],
            )
            basis = basis * pseudoscalar.unsqueeze(1)

        return basis  # [n_atoms, 2*o3_lambda+1, 2*o3_lambda+1]


def cg_combine(A, B, C):
    return torch.einsum("im, in, mnp-> ip", A, B, C)


def get_cg_coefficients(l_max):
    cg_object = ClebschGordanReal()
    for l1 in range(l_max + 1):
        for l2 in range(l_max + 1):
            for L in range(abs(l1 - l2), min(l1 + l2, l_max) + 1):
                cg_object._add(l1, l2, L)
    return cg_object


class ClebschGordanReal:
    def __init__(self):
        self._cgs = {}

    def _add(self, l1, l2, L):
        # print(f"Adding new CGs with l1={l1}, l2={l2}, L={L}")

        if self._cgs is None:
            raise ValueError("Trying to add CGs when not initialized... exiting")

        if (l1, l2, L) in self._cgs:
            raise ValueError("Trying to add CGs that are already present... exiting")

        maxx = max(l1, max(l2, L))

        # real-to-complex and complex-to-real transformations as matrices
        r2c = {}
        c2r = {}
        for l in range(0, maxx + 1):  # noqa: E741
            r2c[l] = _real2complex(l)
            c2r[l] = np.conjugate(r2c[l]).T

        complex_cg = _complex_clebsch_gordan_matrix(l1, l2, L)

        real_cg = (r2c[l1].T @ complex_cg.reshape(2 * l1 + 1, -1)).reshape(
            complex_cg.shape
        )

        real_cg = real_cg.swapaxes(0, 1)
        real_cg = (r2c[l2].T @ real_cg.reshape(2 * l2 + 1, -1)).reshape(real_cg.shape)
        real_cg = real_cg.swapaxes(0, 1)

        real_cg = real_cg @ c2r[L].T

        if (l1 + l2 + L) % 2 == 0:
            rcg = np.real(real_cg)
        else:
            rcg = np.imag(real_cg)

        # Zero any possible (and very rare) near-zero elements
        where_almost_zero = np.where(
            np.logical_and(np.abs(rcg) > 0, np.abs(rcg) < 1e-14)
        )
        if len(where_almost_zero[0] != 0):
            print("INFO: Found almost-zero CG!")
        for i0, i1, i2 in zip(
            where_almost_zero[0], where_almost_zero[1], where_almost_zero[2]
        ):
            rcg[i0, i1, i2] = 0.0

        self._cgs[(l1, l2, L)] = torch.tensor(rcg)

    def get(self, key):
        if key in self._cgs:
            return self._cgs[key]
        else:
            self._add(key[0], key[1], key[2])
            return self._cgs[key]


def _real2complex(L):
    """
    Computes a matrix that can be used to convert from real to complex-valued
    spherical harmonics(coefficients) of order L.

    It's meant to be applied to the left, ``real2complex @ [-L..L]``.
    """
    result = np.zeros((2 * L + 1, 2 * L + 1), dtype=np.complex128)

    I_SQRT_2 = 1.0 / np.sqrt(2)

    for m in range(-L, L + 1):
        if m < 0:
            result[L - m, L + m] = I_SQRT_2 * 1j * (-1) ** m
            result[L + m, L + m] = -I_SQRT_2 * 1j

        if m == 0:
            result[L, L] = 1.0

        if m > 0:
            result[L + m, L + m] = I_SQRT_2 * (-1) ** m
            result[L - m, L + m] = I_SQRT_2

    return result


def _complex_clebsch_gordan_matrix(l1, l2, L):
    if np.abs(l1 - l2) > L or np.abs(l1 + l2) < L:
        return np.zeros((2 * l1 + 1, 2 * l2 + 1, 2 * L + 1), dtype=np.double)
    else:
        return wigners.clebsch_gordan_array(l1, l2, L)


class FakeVectorBasis(torch.nn.Module):
    # fake class to make torchscript work

    def forward(
        self, systems: List[System], selected_atoms: Optional[Labels]
    ) -> torch.Tensor:
        return torch.tensor(0)
